import numpy as np
import corner
from matplotlib.colors import colorConverter
from matplotlib.patches import Patch as mplpatches
from matplotlib.lines import Line2D


def print_mcmc_params(mcmcParams, quantileLevels):
    '''
    This function is just for printing MCMC parameters for easy copy/pasting
    to alog. Update of Jeff's existing code for prettier formatting.

    Prints in two formats: first the quantile values,
    then in the format 'X (+Y/-Z)'.

    Parameters
    ----------
    mcmcParams: dict
        Keys should be parameter names, values include quantiles, errbars, labels
        TODO improve documentation

    Returns
    -------
    tableQuant: str
        Printable table of values in quantile format
    tablePM: str
        Printable table of values in +/- format

    '''

    # Set up lists to hold the tables (will be joined with newline later)
    tableQuant = []
    tablePM = []

    # Define a column spacer
    spacer = " | "

    # Set up left column width for parameter labels
    ncharsLabel = max([len(p['label']) for p in mcmcParams.values()])

    # Set up widths for the quantiles section
    ncharsCol = max([len(f"{x:4.4g}") for p in mcmcParams.values()
                    for x in p['quantiles']])

    # Set up the header for the quantiles section
    tag = "(quantile)"
    pline = [f"{'Parameter':<{ncharsLabel-len(tag)}s}{tag}"]
    for x in quantileLevels:
        pline += [f"{str(x):<{ncharsCol}s}"]
    header = spacer.join(pline)
    tableQuant += [header]
    tableQuant += ["-"*len(header)]

    # Set up each line for the quantiles section
    for param in mcmcParams.values():
        pline = [f"{param['label']:<{ncharsLabel}s}"]
        for x in param['quantiles']:
            valstr = f"{x:4.4g}"
            pline += [f"{valstr:<{ncharsCol}s}"]
        tableQuant += [spacer.join(pline)]
    tableQuant += ["-"*len(header)]

    # Set up column widths for the +/- section
    ncharsCol = max([len(f"{param[key]:4.4g} ({abs(param[key]/param['median']):.2f}%)")
                    for param in mcmcParams.values()
                    for key in ['median', 'errplus', 'errminus']])

    # Set up head for the +/- section
    tag = "(value +/-)"
    pline = [f"{'Parameter':<{ncharsLabel-len(tag)}s}{tag}"]
    for x in ["value", " +", " -"]:
        pline += [f"{str(x):<{ncharsCol}s}"]
    header = spacer.join(pline)
    tablePM += [header]
    tablePM += ["-"*len(header)]

    # Set up each line for the +/- section
    for param in mcmcParams.values():
        pline = [f"{param['label']:<{ncharsLabel}s}"]
        for key in ['median', 'errplus', 'errminus']:
            if key == 'median':
                fmat = f"{param[key]:4.4g}"
            else:
                fmat = f"{param[key]:4.4g} ({(param[key]/param['median']):.2f}%)"
            pline += [f"{fmat:<{ncharsCol}s}"]
        tablePM += [spacer.join(pline)]

    tablePM = "\n".join(tablePM)
    tableQuant = "\n".join(tableQuant)

    return tableQuant, tablePM


def make_corner_plot(chains, mcmcparams, quantilelevels, outfile,
                     title):
    '''
    Make an MCMC corner plot
    parameters
    ----------
    chains: 2d array
        size: (number of mcmc params),(number of steps in mcmc chain)

    mcmcparams: dict
        keys should be parameter names, values include quantiles, errbars, labels

    quantilelevels: 1d array

    outfile
        path to save corner plot

    title
        plot title

    returns
    -------
    none
        just saves figures
    '''

    color = 'C3'  # base fill color for contours
    truthcolor = 'C0'  # color for median markers
    nbins = 100  # number of bins for histogram

    # === create main plot
    cp = corner.corner(
        np.transpose(chains),
        bins=nbins,
        quantiles=[quantilelevels[0], quantilelevels[-1]],
        smooth=2.0,
        labels=[param['mathlabel'] for param in mcmcparams.values()],
        verbose=False,
        label_kwargs={'fontsize': 12},
        show_titles=True,
        title_fmt='.3e',
        title_kwargs={'fontsize': 12},
        plot_datapoints=False,
        plot_density=True,
        plot_contours=True,
        fill_contours=True,
        color=color,
        use_math_text=True,
        truths=[param['median'] for param in mcmcparams.values()],
        truth_color=truthcolor
    )
    cp.suptitle(title, fontsize=20)

    # ==== reproducing the contour colors
    # annoying as this is, corner does not provide a way to get
    # the contour fill color maps, nor the default levels.
    # we copy the following from its definition in
    # https://github.com/dfm/corner.py/blob/e65dd4cdeb7a9f7f75cbcecb4f07a07de65e2cea/src/corner/core.py
    levels = 1.0 - np.exp(-0.5 * np.arange(0.5, 2.1, 0.5) ** 2)
    rgba_color = colorConverter.to_rgba(color)
    contour_cmap = [list(rgba_color) for lev in levels] + [rgba_color]
    for i, l in enumerate(levels):
        contour_cmap[i][-1] *= float(i) / (len(levels) + 1)
    contour_cmap = contour_cmap[-3:][::-1]

    # === generating custom legend

    # shaded rectangle for each contour fill color
    legend_symbols = [mplpatches(
        facecolor=contour_cmap[i],
        edgecolor=contour_cmap[i]) for i in range(len(contour_cmap))]
    # line for the "truth" values (mcmcparam median values)
    legend_symbols.append(Line2D([0], [0], color=truthcolor, lw=3))
    # empty rectangle to make space for bin count label
    legend_symbols.append(mplpatches(alpha=0))

    # create legend
    cp.legend(
        legend_symbols,
        [r'$1\sigma$',
         r'$2\sigma$',
         r'$3\sigma$',
         'map',
         f'({nbins} bins for 1d pdf)'],
        fontsize=15,
        title_fontsize=15,
        title="2d pdf contours",
        frameon=True,
        markerscale=20.0,
    )

    # === fix up the histogram axes

    # grab all subplot axes
    axes = cp.get_axes()

    # determine which ones are histograms
    nparams = len(mcmcparams.values())
    length_sides = np.arange(nparams)
    histogram_indices = (nparams+1)*length_sides

    # create right-side axis for each histogram
    for i in histogram_indices:
        ax = axes[i]
        ax2 = ax.twinx()
        ax2.set_ylabel('1d norm. pdf \n (percent per bin)')
        axes += [ax2]
        ax2.set_ylim(0, 1.1)

    # resize the tick params to make them smaller
    for ax in axes:
        ax.tick_params(axis='both', labelsize=10)

    # save figure
    cp.set_size_inches(10, 10)
    cp.tight_layout()
    cp.savefig(outfile)
    return cp



import pydarm
import numpy as np
# import matplotlib.pyplot as plt
"""
This script gives a few example usages of the plot method pydarm/plot.py.
The method tries to be flexible with the input. See the code for all
possibilities.
The plot functions below DO NOT produce an interacive graph unless you
run plt.show()
"""

# --------------------------------------------------------------------
# Example with two darm models (there are other ways to make 2 models)

config = '../example_model_files/myH1_20190416.ini'

D1 = pydarm.darm.DARMModel(config)
D2 = pydarm.darm.DARMModel(config)
D2.sensing.coupled_cavity_pole_frequency = 150

# You can simply do the following to get the critique plots.
D1.plot(filename='plot_example_critique_1.pdf', label=['H1'])

# Do the following if you want to get critique plots and at the same
# time compare two models.
pydarm.plot.critique(D1, D2, label=['Normal D1', 'Crazy D2'],
                     filename='plot_example_critique_2.pdf', ifo='H1', ugf_start=10, ugf_end=100)

# If you don't want all the plots but just one kind do this
pydarm.plot.critique(D1, plot_selection='olg', label=['OLG'], filename='plot_example_just_olg.pdf')

# --------------------------------------------------------------------
# Example with different kinds of inputs. The data plotted are duplicates
# but that shouldn't matter for this demonstration.

# Frequency list for a freq, tf input.
freq = np.logspace(-1, 4, 1000)
tf = D1.compute_darm_olg(freq)

# Create a tuple to show that this works too for plotting
test_tuple2 = (freq, tf)

# Example of inputting data, to see that error bars are taken care of.

# These are the xml files we want to get our data from.
measurement_file_1 = \
    '../test/2020-01-03_H1_DARM_OLGTF_LF_SS_5to1100Hz_15min.xml'

meas_obj_1 = pydarm.measurement.Measurement(measurement_file_1)
channelA_1 = 'H1:LSC-DARM1_IN2'
channelB_1 = 'H1:LSC-DARM1_IN1'

measurement_tuple = meas_obj_1.get_raw_tf(channelA_1,
                                          channelB_1,
                                          cohThresh=0.0)
pydarm.plot.residuals(freq, -tf,
                      measurement_tuple[0], measurement_tuple[1], measurement_tuple[3],
                      mag_min=0.8, mag_max=1.2, phase_min=-5, phase_max=5,
                      filename='plot_example_residuals.pdf', plot_input=True, show=False)

# if you really want to save code, just input a method.
pydarm.plot.plot(D1.compute_darm_olg, label=['method'], filename='plot_example_method.pdf',
                 show=False, greed=True)

pydarm.plot.plot(test_tuple2,
                 freq, D1.compute_darm_olg(freq),
                 freq, D1.compute_darm_olg(freq)*2,
                 measurement_tuple,
                 title='Test Plot',
                 label=['2 tuple', '1+1', '1+1', '4 tuple'],
                 style='example.mplstyle',
                 filename='plot_example_multiinputs.pdf')
# plt.show()



from ._util import logger


##################################################

def add_args(parser):
    pass


def main(args):
    """show the raw data plots from lastest measurements

    """
    logger.error("not implemented!")



import unittest
import pydarm
import numpy as np


class TestPcalDewhitenResponse(unittest.TestCase):

    def setUp(self):
        # Pre-computed values
        # frequencies = np.logspace(0, np.log10(5000.), 10)
        # pcal_dewhiten = 1.0, 1.0
        self.frequencies = np.logspace(0, np.log10(5000.), 10)
        self.known_pcal_dewhiten_response = np.array(
            [0-0.5j,
             -0.09664744749943081-0.08833721090114179j,
             -0.021210239147445686-0.0065396429418788145j,
             -0.00338506272710137-0.00039727801004293923j,
             -0.0005144644812257184-2.3368054441227016e-05j,
             -7.7612591723805e-05-1.3677683875486927e-06j,
             -1.169566056662415e-05-7.999812866147876e-08j,
             -1.7621560448865418e-06-4.678411892855308e-09j,
             -2.654929417882733e-07-2.7359600619251967e-10j,
             -3.999999520000034e-08-1.5999998720000092e-11j])

    def tearDown(self):
        del self.frequencies
        del self.known_pcal_dewhiten_response

    def test_pcal_dewhiten_response(self):
        """ Test the dewhitening filter response """
        pcal = pydarm.pcal.PcalModel('''
[metadata]
[interferometer]
[pcal]
pcal_dewhiten = 1.0, 1.0
''')
        dewhitening_response = pcal.pcal_dewhiten_response(self.frequencies)
        for n in range(len(self.frequencies)):
            self.assertAlmostEqual(
                np.abs(dewhitening_response[n]),
                np.abs(self.known_pcal_dewhiten_response[n]))
            self.assertAlmostEqual(
                np.angle(dewhitening_response[n], deg=True),
                np.angle(self.known_pcal_dewhiten_response[n], deg=True))


class TestComputePcalCorrection(unittest.TestCase):

    def setUp(self):
        # Pre-computed values
        self.frequencies = np.logspace(0, np.log10(5000.), 10)
        self.model_string = '''
[metadata]
[interferometer]
[pcal]
pcal_dewhiten = 1.0, 1.0
ref_pcal_2_darm_act_sign = -1
analog_anti_aliasing_file = test/H1aa.mat
anti_aliasing_rate_string = 16k
anti_aliasing_method = biquad
'''
        self.known_pcal_correction_no_delay = np.array(
            [-0.00025358893371024303+0.49999958238994285j,
             0.09653186519726428+0.08846335096661345j,
             0.021188078968294916+0.0066109980202688125j,
             0.003381476169995326+0.0004266184275991344j,
             0.000513802030855575+3.485552121587564e-05j,
             7.739366982337172e-05+5.8303787826440386e-06j,
             1.1543776066542702e-05+1.806974404649834e-06j,
             1.6213496404216842e-06+6.607708336536615e-07j,
             1.329595567674792e-07+2.1863432594589902e-07j,
             -4.0222497230234286e-08+2.2455753017906737e-09j])
        self.known_pcal_correction = np.array(
            [-0.0004453363486970135+0.49999944837269694j,
             0.09644441637641843+0.08855868118099472j,
             0.02117118282116747+0.006664908337209392j,
             0.0033786058580479697+0.0004487837195992806j,
             0.0005131398661356246+4.353058755469486e-05j,
             7.706668186793886e-05+9.192387684605753e-06j,
             1.1269075176834982e-05+3.0873721888339655e-06j,
             1.3659131088236492e-06+1.0952963663988825e-06j,
             -5.030899158708751e-08+2.508948337114827e-07j,
             1.1554697588642775e-08-3.8592497399199635e-08j])

    def tearDown(self):
        del self.frequencies
        del self.model_string
        del self.known_pcal_correction_no_delay
        del self.known_pcal_correction

    def test_compute_pcal_correction_no_delay(self):
        """ Test the Pcal correction at the end station"""
        pcal = pydarm.pcal.PcalModel(self.model_string)
        pcal_correction = pcal.compute_pcal_correction(self.frequencies, True)
        for n in range(len(self.frequencies)):
            self.assertAlmostEqual(
                np.abs(pcal_correction[n]),
                np.abs(self.known_pcal_correction_no_delay[n]))
            self.assertAlmostEqual(
                np.angle(pcal_correction[n], deg=True),
                np.angle(self.known_pcal_correction_no_delay[n], deg=True))

    def test_compute_pcal_correction(self):
        """ Test the Pcal correction at the corner station"""
        pcal = pydarm.pcal.PcalModel(self.model_string)
        pcal_correction = pcal.compute_pcal_correction(self.frequencies, False)
        for n in range(len(self.frequencies)):
            self.assertAlmostEqual(
                np.abs(pcal_correction[n]),
                np.abs(self.known_pcal_correction[n]))
            self.assertAlmostEqual(
                np.angle(pcal_correction[n], deg=True),
                np.angle(self.known_pcal_correction[n], deg=True))

class TestDigitalFilterResponse(unittest.TestCase):

    def setUp(self):
        # Pre-computed values
        # frequencies = np.logspace(0, np.log10(5000.), 10)
        self.frequencies = np.logspace(0, np.log10(5000.), 10)
        self.known_digital_filter_response = np.array(
            [0.003032834984214-0.000852215454730j,
             0.000583953846217-0.000533208471501j,
             0.000630611562557-0.000194421040491j,
             0.000637732123677-0.000074841527594j,
             0.000638802551708-0.000029013638555j,
             0.000638963766980-0.000011258145407j,
             0.000638988054812-0.000004365874356j,
             0.000638991714044-0.000001684584793j,
             0.000638992265247-0.000000627782359j,
             0.000638992347469-0.000000172013907j])

    def tearDown(self):
        del self.frequencies
        del self.known_digital_filter_response

    def test_digital_filter_response(self):
        """ Test the digital filter response """
        pcal = pydarm.pcal.PcalModel('''
[metadata]
[interferometer]
[pcal]
pcal_filter_file           = test/H1CALEY_1123041152.txt
pcal_filter_bank           = PCALY_TX_PD
pcal_filter_modules_in_use = 6,8
pcal_filter_gain           = 1.0
''')
        digital_filter = pcal.digital_filter_response(self.frequencies)
        for n in range(len(self.frequencies)):
            self.assertAlmostEqual(
                np.abs(digital_filter[n]),
                np.abs(self.known_digital_filter_response[n]))
            self.assertAlmostEqual(
                np.angle(digital_filter[n], deg=True),
                np.angle(self.known_digital_filter_response[n], deg=True), places=6)

if __name__ == '__main__':
    unittest.main()



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join('..', 'pydarm')))

from pydarm import __version__


# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'pyDARM'
copyright = '2022, Evan Goetz'
author = 'Evan Goetz'
if "dev" in __version__:
    release = version = "dev"
else:
    release = version = __version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

default_role = "obj"
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'numpydoc',
    'sphinx_immaterial_igwn',
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_immaterial_igwn'
html_static_path = ['_static']
html_theme_options = {
    # metadata
    "edit_uri": "blob/main/docs",
    "repo_name": "pyDARM",
    "repo_type": "gitlab",
    "repo_url": "https://git.ligo.org/Calibration/pydarm",
    "icon": {
        "repo": "fontawesome/brands/gitlab",
        "edit": "material/file-edit-outline",
    },
    "features": [
        "navigation.sections",
    ],
    # colouring and light/dark mode
    "palette": [
        {
            "media": "(prefers-color-scheme: light)",
            "scheme": "default",
            "primary": "blue-grey",
            "accent": "deep-orange",
            "toggle": {
                "icon": "material/eye-outline",
                "name": "Switch to dark mode",
            },
        },
        {
            "media": "(prefers-color-scheme: dark)",
            "scheme": "slate",
            "primary": "amber",
            "accent": "deep-orange",
            "toggle": {
                "icon": "material/eye",
                "name": "Switch to light mode",
            },
        },
    ],
    # table of contents
    "toc_title_is_page_title": True,
}

# -- autodoc

autoclass_content = 'class'
autodoc_default_flags = ['show-inheritance', 'members', 'inherited-members']

# -- autosummary

autosummary_generate = True

# -- numpydoc

# fix numpydoc autosummary
numpydoc_show_class_members = False

# use blockquotes (numpydoc>=0.8 only)
numpydoc_use_blockquotes = True



# Copyright (C) Evan Goetz (2021)
#
# This file is part of pyDARM.
#
# pyDARM is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pyDARM is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pyDARM. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from scipy import signal

from .utils import compute_digital_filter_response, digital_delay_filter
from .model import Model


class PcalModel(Model):
    """
    A photon calibrator (pcal) model object

    The class serves to return transfer functions that
    are important to the pcal system, be it corrections
    to the response when pulling the PCAL channels from
    the frames, or establishing force coefficients.

    """

    def __init__(self, config):
        super().__init__(config, measurement='pcal')

    def pcal_dewhiten_response(self, frequencies):
        """
        Compute the dewhitening response

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        tf : `complex128`, array-like
            transfer function response of the Pcal dewhitening filter

        """

        filt = signal.ZerosPolesGain(
            [], -2.0*np.pi*np.asarray(self.pcal_dewhiten),
            np.prod(-2.0*np.pi*np.asarray(self.pcal_dewhiten)))

        return signal.freqresp(filt, 2.0*np.pi*frequencies)[1]

    def compute_pcal_correction(self, frequencies, endstation=False,
                                include_dewhitening=True):
        """
        Compute the Pcal correction for the offline analysis
        See G1501518
        NOTE: This also includes the Pcal arm sign, so be aware!

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response
        endstation : `bool`, optional
            When false (default), the correction is computed for CAL-CS
            PCAL channel, which includes 1 16k clock cycle delay that we must
            compensate (undo). Otherwise, when true, the correction is computed
            at the end station, which does not include 1 16k clock cycle delay
        include_dewhitening : `bool`, optional
            if the dewhitening filter is on (default), then the correction will
            include the two 1 Hz poles

        Returns
        -------
        tf : `complex128`, array-like
            transfer function response of the correction

        """

        pcal_sign = self.ref_pcal_2_darm_act_sign

        pcal_dewhitening = 1
        if include_dewhitening:
            pcal_dewhitening = self.pcal_dewhiten_response(frequencies)

        pcal_analog_aa = self.analog_aa_or_ai_filter_response(frequencies)
        # As per the Pcal correction factor shown in G1501518, the front end
        # Pcal calibration already takes into account any analog AA gain
        # (if there is any). Therefore, we don't need the analog AA gain
        # to be included in the pcal correction factor
        pcal_analog_aa_hi_f = pcal_analog_aa / \
            np.abs(self.analog_aa_or_ai_filter_response(1e-2))

        pcal_digital_aa = self.digital_aa_or_ai_filter_response(frequencies)

        # delay filter for end station to CALCS otherwise no delay is used
        # in the correction factor
        if endstation is False:
            end_to_calcs_response = (
                signal.dfreqresp(digital_delay_filter(1, 2**14),
                                 2.0*np.pi*frequencies/2**14)[1])
        else:
            end_to_calcs_response = 1

        return (pcal_sign *
                pcal_dewhitening /
                (pcal_analog_aa_hi_f * pcal_digital_aa) /
                end_to_calcs_response)

    def digital_filter_response(self, frequencies):
        """
        Importing suspension filter from FOTON file which normalized by whitening
          filter 2 1Hz poles (susnorm) and mpN_DC gain.

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        tf : `complex128`, array-like
            transfer function response of the correction

        """

        response = np.ones(len(frequencies), dtype='complex128')
        for n in range(len(self.pcal_filter_modules_in_use)):
            if n == 0:
                tf, pfilt = compute_digital_filter_response(
                    self.dpath(self.pcal_filter_file),
                    self.pcal_filter_bank,
                    self.pcal_filter_modules_in_use[n],
                    self.pcal_filter_gain, frequencies, pfilt=None)
            else:
                tf = compute_digital_filter_response(
                    self.dpath(self.pcal_filter_file),
                    self.pcal_filter_bank,
                    self.pcal_filter_modules_in_use[n],
                    self.pcal_filter_gain, frequencies, pfilt=pfilt)[0]
            response *= tf

        return response



from ._util import logger
from ._util import CMDError
from .measurement_set import MeasurementSet
from .measurement_set import list_all_measurement_sets


CMDS = (
    'env',
    'config',
    'model',
    'ls',
    'measure',
    'show',
    'report',
    'uncertainty',
    'status',
    'export',
)



import unittest
import pydarm
import numpy as np


class TestComputeDigitalFilterResponse(unittest.TestCase):

    def setUp(self):
        # frequencies = np.logspace(0, np.log10(5000.), 10)
        self.frequencies = np.logspace(0, np.log10(5000.), 10)
        self.known_digital_filter_response = np.array(
            [-3513076500191.6504+3327250871150.924j,
             -164538201540.2522+209970221596.44916j,
             -3242877524.3140893-646390346.0757524j,
             3152750901.033689+60862553.6703552j,
             5968957387.406848+6106710482.173726j,
             17063001608.677845+14901038698.213606j,
             59173157596.065796-25364518140.6575j,
             -46407563930.73129+21858627480.823456j,
             -9021881.73649469-6742239.626013521j,
             -27770965.672582164+52593450.2554684j])

    def tearDown(self):
        del self.frequencies
        del self.known_digital_filter_response

    def test_compute_digital_filter_response(self):
        """ Test the DARM digital filter repsonse """
        D = pydarm.darm.DigitalModel('''
[metadata]
[interferometer]
[digital]
digital_filter_file = test/H1OMC_1239468752.txt
digital_filter_bank = LSC_DARM1, LSC_DARM2
digital_filter_modules = 1,2,3,4,7,9,10: 3,4,5,6,7
digital_filter_gain = 400,1
''')
        test_response = D.compute_response(
            self.frequencies)
        for n in range(len(self.frequencies)):
            # Requires investigation why this delta tolerance has to be higher
            # on some systems in order to pass.
            # Can test using print() to check output and call test using
            # pytest -s test/darm_test.py
            self.assertAlmostEqual(
                np.abs(test_response[n]) /
                np.abs(self.known_digital_filter_response[n]), 1.0)
            self.assertAlmostEqual(
                np.angle(test_response[n], deg=True) -
                np.angle(self.known_digital_filter_response[n], deg=True), 0.0,
                places=5)


class TestComputeDarmOlg(unittest.TestCase):

    def setUp(self):
        # frequencies = np.logspace(0, np.log10(5000.), 10)
        self.frequencies = np.logspace(0, np.log10(5000.), 10)
        self.known_olg = np.array(
            [-346589.7289879694+67697.98179370671j,
             -1859.4410636292505+7152.898242047175j,
             40.79712693912431-10.496083469022562j,
             -3.200667505597764+0.6749176141222187j,
             -1.0964382128049217-0.6352470027806889j,
             -0.48849524504067693-0.12453097698086546j,
             0.07580387098963454+0.20189845660408226j,
             -0.010730361938738586+0.010017089450759862j,
             1.1367097388995271e-07-1.7573024415341907e-07j,
             -1.8569508267009735e-08+2.419265648650538e-08j])

    def tearDown(self):
        del self.frequencies
        del self.known_olg

    def test_compute_darm_olg(self):
        """ Test DARM open loop gain response """
        darm = pydarm.darm.DARMModel('''
[sensing]
x_arm_length = 3994.4704
y_arm_length = 3994.4692
coupled_cavity_optical_gain = 3.22e6
coupled_cavity_pole_frequency = 410.6
detuned_spring_frequency = 4.468
detuned_spring_Q = 52.14
sensing_sign = 1
is_pro_spring = True
anti_aliasing_rate_string = 16k
anti_aliasing_method      = biquad
analog_anti_aliasing_file = test/H1aa.mat, test/H1aa.mat
omc_meas_p_trans_amplifier_uncompensated   = 13.7e3, 17.8e3: 13.7e3, 17.8e3
whitening_mode_names = test, test
omc_meas_p_whitening_uncompensated_test   = 11.346e3, 32.875e3, 32.875e3: 11.521e3, 32.863e3, 32.863e3
super_high_frequency_poles_apparent_delay = 0, 0
gain_ratio = 1, 1
balance_matrix = 1, 1
omc_path_names = A, B
single_pole_approximation_delay_correction = -12e-6
adc_gain = 1638.001638001638, 1638.001638001638
omc_filter_file = test/H1OMC_1239468752.txt
omc_filter_bank = OMC_DCPD_A, OMC_DCPD_B
omc_filter_noncompensating_modules = 4: 4
omc_filter_gain = 1, 1
omc_front_end_trans_amplifier_compensation = ON, ON
omc_front_end_whitening_compensation_test = ON, ON

[actuation_x_arm]
darm_feedback_sign = -1
tst_NpV2 = 4.427e-11
linearization = OFF
actuation_esd_bias_voltage = -9.3
suspension_file = test/H1susdata_O3.mat
tst_driver_uncompensated_Z_UL = 129.7e3
tst_driver_uncompensated_Z_LL = 90.74e3
tst_driver_uncompensated_Z_UR = 93.52e3
tst_driver_uncompensated_Z_LR = 131.5e3
tst_driver_uncompensated_P_UL = 3.213e3, 31.5e3
tst_driver_uncompensated_P_LL = 3.177e3, 26.7e3
tst_driver_uncompensated_P_UR = 3.279e3, 26.6e3
tst_driver_uncompensated_P_LR = 3.238e3, 31.6e3
tst_driver_DC_gain_VpV_HV = 40
tst_driver_DC_gain_VpV_LV = 1.881
anti_imaging_rate_string = 16k
anti_imaging_method = biquad
analog_anti_imaging_file = test/H1aa.mat
dac_gain = 7.62939453125e-05
unknown_actuation_delay = 15e-6
pum_driver_DC_trans_ApV = 2.6847e-4
pum_coil_outf_signflip = 1
pum_NpA = 0.02947
uim_driver_DC_trans_ApV = 6.1535e-4
uim_NpA = 1.634
sus_filter_file = test/H1SUSETMX_1236641144.txt
tst_isc_inf_bank = ETMX_L3_ISCINF_L
tst_isc_inf_modules =
tst_isc_inf_gain = 1.0
tst_lock_bank = ETMX_L3_LOCK_L
tst_lock_modules = 5,8,9,10
tst_lock_gain = 1.0
tst_drive_align_bank = ETMX_L3_DRIVEALIGN_L2L
tst_drive_align_modules = 4,5
tst_drive_align_gain = -35.7
pum_lock_bank = ETMX_L2_LOCK_L
pum_lock_modules = 7
pum_lock_gain = 23.0
pum_drive_align_bank = ETMX_L2_DRIVEALIGN_L2L
pum_drive_align_modules = 6,7
pum_drive_align_gain = 1.0
uim_lock_bank = ETMX_L1_LOCK_L
uim_lock_modules = 10
uim_lock_gain = 1.06
uim_drive_align_bank = ETMX_L1_DRIVEALIGN_L2L
uim_drive_align_modules =
uim_drive_align_gain = 1.0

[actuation]
darm_output_matrix = 1.0, -1.0, 0.0, 0.0
darm_feedback_x = OFF, ON, ON, ON
darm_feedback_y = OFF, OFF, OFF, OFF

[digital]
digital_filter_file = test/H1OMC_1239468752.txt
digital_filter_bank = LSC_DARM1, LSC_DARM2
digital_filter_modules = 1,2,3,4,7,9,10: 3,4,5,6,7
digital_filter_gain = 400,1
''')
        test_olg = darm.compute_darm_olg(self.frequencies)
        for n in range(len(self.frequencies)):
            self.assertAlmostEqual(
                np.abs(test_olg[n]) / np.abs(self.known_olg[n]), 1.0)
            self.assertAlmostEqual(
                np.angle(test_olg[n], deg=True) -
                np.angle(self.known_olg[n], deg=True), 0.0, places=5)


class TestComputeResponseFunction(unittest.TestCase):

    def setUp(self):
        # frequencies = np.logspace(0, np.log10(5000.), 10)
        self.frequencies = np.logspace(0, np.log10(5000.), 10)
        self.known_response = np.array(
            [2.0406181004597967-0.4013213677915312j,
             0.0011220990463462926-0.004469979302356652j,
             7.174381300234404e-06-1.4617952572908577e-06j,
             -6.479990333472527e-07+1.5653156901534748e-07j,
             -1.3662040618868035e-09-1.9861756619000937e-07j,
             1.6810351811864666e-07+2.0548446256238736e-08j,
             2.1237360469410807e-07+3.590194203666282e-07j,
             -7.07952406495893e-08+6.371681259820066e-07j,
             -1.4643321723110076e-06+3.0336488559992614e-07j,
             4.465169869308327e-06-1.5260050031500522e-06j])

    def tearDown(self):
        del self.frequencies
        del self.known_response

    def test_compute_darm_response_function(self):
        """ Test DARM closed loop response function """
        model_string = '''
[metadata]
[interferometer]
[sensing]
x_arm_length = 3994.4704
y_arm_length = 3994.4692
coupled_cavity_optical_gain = 3.22e6
coupled_cavity_pole_frequency = 410.6
detuned_spring_frequency = 4.468
detuned_spring_Q = 52.14
sensing_sign = 1
is_pro_spring = True
anti_aliasing_rate_string = 16k
anti_aliasing_method      = biquad
analog_anti_aliasing_file = test/H1aa.mat, test/H1aa.mat
omc_meas_p_trans_amplifier_uncompensated   = 13.7e3, 17.8e3: 13.7e3, 17.8e3
whitening_mode_names = test, test
omc_meas_p_whitening_uncompensated_test   = 11.346e3, 32.875e3, 32.875e3: 11.521e3, 32.863e3, 32.863e3
super_high_frequency_poles_apparent_delay = 0, 0
gain_ratio = 1, 1
balance_matrix = 1, 1
omc_path_names = A, B
single_pole_approximation_delay_correction = -12e-6
adc_gain = 1, 1
omc_front_end_trans_amplifier_compensation = ON, ON
omc_front_end_whitening_compensation_test = ON, ON

[actuation_x_arm]
darm_feedback_sign = -1
tst_NpV2 = 4.427e-11
linearization = OFF
actuation_esd_bias_voltage = -9.3
suspension_file = test/H1susdata_O3.mat
tst_driver_uncompensated_Z_UL = 129.7e3
tst_driver_uncompensated_Z_LL = 90.74e3
tst_driver_uncompensated_Z_UR = 93.52e3
tst_driver_uncompensated_Z_LR = 131.5e3
tst_driver_uncompensated_P_UL = 3.213e3, 31.5e3
tst_driver_uncompensated_P_LL = 3.177e3, 26.7e3
tst_driver_uncompensated_P_UR = 3.279e3, 26.6e3
tst_driver_uncompensated_P_LR = 3.238e3, 31.6e3
tst_driver_DC_gain_VpV_HV = 40
tst_driver_DC_gain_VpV_LV = 1.881
anti_imaging_rate_string = 16k
anti_imaging_method = biquad
analog_anti_imaging_file = test/H1aa.mat
dac_gain = 7.62939453125e-05
unknown_actuation_delay = 15e-6
pum_driver_DC_trans_ApV = 2.6847e-4
pum_coil_outf_signflip = 1
pum_NpA = 0.02947
uim_driver_DC_trans_ApV = 6.1535e-4
uim_NpA = 1.634
sus_filter_file = test/H1SUSETMX_1236641144.txt
tst_isc_inf_bank = ETMX_L3_ISCINF_L
tst_isc_inf_modules =
tst_isc_inf_gain = 1.0
tst_lock_bank = ETMX_L3_LOCK_L
tst_lock_modules = 5,8,9,10
tst_lock_gain = 1.0
tst_drive_align_bank = ETMX_L3_DRIVEALIGN_L2L
tst_drive_align_modules = 4,5
tst_drive_align_gain = -35.7
pum_lock_bank = ETMX_L2_LOCK_L
pum_lock_modules = 7
pum_lock_gain = 23.0
pum_drive_align_bank = ETMX_L2_DRIVEALIGN_L2L
pum_drive_align_modules = 6,7
pum_drive_align_gain = 1.0
uim_lock_bank = ETMX_L1_LOCK_L
uim_lock_modules = 10
uim_lock_gain = 1.06
uim_drive_align_bank = ETMX_L1_DRIVEALIGN_L2L
uim_drive_align_modules =
uim_drive_align_gain = 1.0

[actuation]
darm_output_matrix = 1.0, -1.0, 0.0, 0.0
darm_feedback_x = OFF, ON, ON, ON
darm_feedback_y = OFF, OFF, OFF, OFF

[digital]
digital_filter_file = test/H1OMC_1239468752.txt
digital_filter_bank = LSC_DARM1, LSC_DARM2
digital_filter_modules = 1,2,3,4,7,9,10: 3,4,5,6,7
digital_filter_gain = 400,1
'''
        darm = pydarm.darm.DARMModel(model_string)
        test_response = darm.compute_response_function(self.frequencies)
        for n in range(len(self.frequencies)):
            self.assertAlmostEqual(
                np.abs(test_response[n]) / np.abs(self.known_response[n]), 1.0)
            self.assertAlmostEqual(
                np.angle(test_response[n], deg=True) -
                np.angle(self.known_response[n], deg=True), 0.0, places=5)


class TestComputeEtas(unittest.TestCase):

    def setUp(self):
        self.frequencies = np.logspace(np.log10(20), np.log10(2000), 10)
        self.syserr_input = np.array(
            [1.0000003831992847-0.0007929858798234665j,
             0.9999986255804106-0.002042969515282433j,
             0.9999869597043107-0.0052632862823899385j,
             0.9999095302557972-0.013559488366496989j,
             0.9993956411093269-0.03492779907418393j,
             0.995986356934377-0.08988999137528689j,
             0.9734269429695694-0.22996802633705743j,
             0.8267642616923517-0.5649608162253222j,
             -0.004181642663901519-1.003311656263212j,
             -0.19691508479464048+0.7796521952236531j])
        self.known_etas = np.array(
            [[0.99997492-4.43808402e-04j, 0.99816941-1.32625955e-03j,
              0.99042173+2.75217421e-03j, 0.98764461+2.40563425e-02j,
              1.00478583+5.90395517e-02j, 1.01828258+8.71417774e-02j,
              0.94939253+2.12468056e-01j, 0.82362819+5.75180328e-01j,
              -0.00435558+9.98282271e-01j, -0.30462374-1.20565509e+00j],
             [1.-0.00000000e+00j, 1.-0.00000000e+00j,
              1.-0.00000000e+00j, 1.+0.00000000e+00j,
              1.+0.00000000e+00j, 1.-6.08247505e-17j,
              1.+0.00000000e+00j, 1.+0.00000000e+00j,
              1.+0.00000000e+00j, 1.-0.00000000e+00j],
             [0.99997492-4.43808402e-04j, 0.99816941-1.32625955e-03j,
              0.99042173+2.75217421e-03j, 0.98764461+2.40563425e-02j,
              1.00478583+5.90395517e-02j, 1.01828258+8.71417774e-02j,
              0.94939253+2.12468056e-01j, 0.82362819+5.75180328e-01j,
              -0.00435558+9.98282271e-01j, -0.30462374-1.20565509e+00j]])

    def tearDown(self):
        del self.frequencies
        del self.syserr_input
        del self.known_etas

    def test_compute_etas(self):
        model_string = '''
[metadata]
[interferometer]
[sensing]
x_arm_length = 3994.4704
y_arm_length = 3994.4692
coupled_cavity_optical_gain = 3.22e6
coupled_cavity_pole_frequency = 410.6
detuned_spring_frequency = 4.468
detuned_spring_Q = 52.14
sensing_sign = 1
is_pro_spring = True
anti_aliasing_rate_string = 16k
anti_aliasing_method      = biquad
analog_anti_aliasing_file = test/H1aa.mat, test/H1aa.mat
omc_meas_p_trans_amplifier_uncompensated   = 13.7e3, 17.8e3: 13.7e3, 17.8e3
whitening_mode_names = test, test
omc_meas_p_whitening_uncompensated_test   = 11.346e3, 32.875e3, 32.875e3: 11.521e3, 32.863e3, 32.863e3
super_high_frequency_poles_apparent_delay = 0, 0
gain_ratio = 1, 1
balance_matrix = 1, 1
omc_path_names = A, B
single_pole_approximation_delay_correction = -12e-6
adc_gain = 1, 1
omc_front_end_trans_amplifier_compensation = ON, ON
omc_front_end_whitening_compensation_test = ON, ON

[actuation_x_arm]
darm_feedback_sign = -1
tst_NpV2 = 4.427e-11
linearization = OFF
actuation_esd_bias_voltage = -9.3
suspension_file = test/H1susdata_O3.mat
tst_driver_uncompensated_Z_UL = 129.7e3
tst_driver_uncompensated_Z_LL = 90.74e3
tst_driver_uncompensated_Z_UR = 93.52e3
tst_driver_uncompensated_Z_LR = 131.5e3
tst_driver_uncompensated_P_UL = 3.213e3, 31.5e3
tst_driver_uncompensated_P_LL = 3.177e3, 26.7e3
tst_driver_uncompensated_P_UR = 3.279e3, 26.6e3
tst_driver_uncompensated_P_LR = 3.238e3, 31.6e3
tst_driver_DC_gain_VpV_HV = 40
tst_driver_DC_gain_VpV_LV = 1.881
anti_imaging_rate_string = 16k
anti_imaging_method = biquad
analog_anti_imaging_file = test/H1aa.mat
dac_gain = 7.62939453125e-05
unknown_actuation_delay = 15e-6
pum_driver_DC_trans_ApV = 2.6847e-4
pum_coil_outf_signflip = 1
pum_NpA = 0.02947
uim_driver_DC_trans_ApV = 6.1535e-4
uim_NpA = 1.634
sus_filter_file = test/H1SUSETMX_1236641144.txt
tst_isc_inf_bank = ETMX_L3_ISCINF_L
tst_isc_inf_modules =
tst_isc_inf_gain = 1.0
tst_lock_bank = ETMX_L3_LOCK_L
tst_lock_modules = 5,8,9,10
tst_lock_gain = 1.0
tst_drive_align_bank = ETMX_L3_DRIVEALIGN_L2L
tst_drive_align_modules = 4,5
tst_drive_align_gain = -35.7
pum_lock_bank = ETMX_L2_LOCK_L
pum_lock_modules = 7
pum_lock_gain = 23.0
pum_drive_align_bank = ETMX_L2_DRIVEALIGN_L2L
pum_drive_align_modules = 6,7
pum_drive_align_gain = 1.0
uim_lock_bank = ETMX_L1_LOCK_L
uim_lock_modules = 10
uim_lock_gain = 1.06
uim_drive_align_bank = ETMX_L1_DRIVEALIGN_L2L
uim_drive_align_modules =
uim_drive_align_gain = 1.0

[actuation]
darm_output_matrix = 1.0, -1.0, 0.0, 0.0
darm_feedback_x = OFF, ON, ON, ON
darm_feedback_y = OFF, OFF, OFF, OFF

[digital]
digital_filter_file = test/H1OMC_1239468752.txt
digital_filter_bank = LSC_DARM1, LSC_DARM2
digital_filter_modules = 1,2,3,4,7,9,10: 3,4,5,6,7
digital_filter_gain = 400,1
'''
        darm = pydarm.darm.DARMModel(model_string)
        test_eta_R_c, test_eta_R_a, test_eta_R = \
            darm.compute_etas(self.frequencies, sensing_syserr=self.syserr_input)
        for n in range(len(self.frequencies)):
            self.assertAlmostEqual(
                np.abs(test_eta_R_c[n]) / np.abs(self.known_etas[0][n]), 1.0)
            self.assertAlmostEqual(
                np.angle(test_eta_R_c[n], deg=True) -
                np.angle(self.known_etas[0][n], deg=True), 0.0, places=5)
            self.assertAlmostEqual(
                np.abs(test_eta_R_a[n]) / np.abs(self.known_etas[1][n]), 1.0)
            self.assertAlmostEqual(
                np.angle(test_eta_R_a[n], deg=True) -
                np.angle(self.known_etas[1][n], deg=True), 0.0, places=5)
            self.assertAlmostEqual(
                np.abs(test_eta_R[n]) / np.abs(self.known_etas[2][n]), 1.0)
            self.assertAlmostEqual(
                np.angle(test_eta_R[n], deg=True) -
                np.angle(self.known_etas[2][n], deg=True), 0.0, places=5)


if __name__ == '__main__':
    unittest.main()



import os
import argparse
import datetime
import subprocess

import numpy as np
import corner
from scipy.signal import freqresp
from matplotlib import pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages

from ._const import IFO
from . import _util
from ._util import logger
from . import report_utils as ru
from .measurement_set import MeasurementSet
from ..sensing import SensingModel
from ..plot import BodePlot


def find_closest_pcal(pcal_measurements, mfile):
    """find the closest pcal measurement to the given measurement file

    For each measurement file, the datetime of the measurement is
    extracted from the filename.  The closest pcal measurement and
    it's correspond datetime object are returned.

    """
    mtime = _util.measurement_extract_datetime(mfile)
    last = (None, abs(mtime - datetime.datetime(1, 1, 1)))
    for f in pcal_measurements:
        dt = _util.measurement_extract_datetime(f)
        diff = abs(dt - mtime)
        if diff < last[1]:
            last = (f, diff)
    return last


def run_sensing_MCMC(processedSensing, config, mset_id, C_ref, mset_dir):
    # Run the MCMC chain
    mcmcChains = np.transpose(
        processedSensing.run_mcmc(
            fmin=config['mcmc_fmin'],
            fmax=config['mcmc_fmax'],
            burn_in_steps=config['mcmc_burn_in'],
            steps=config['mcmc_steps'],
            save_chain_to_file=os.path.join(
                mset_dir, 'sensing_mcmc_chain.hdf5'),
            save_map_to_file=os.path.join(
                mset_dir, 'sensing_mcmc_map.json'),
        ))

    # Set up a dict of dicts to hold the mcmc parameter information
    mcmcParams = {
        'opticalGain': {
            'label': 'Optical gain, H_c (ct/m)',
            'mathlabel': r'$H_C$'},
        'cavityPole': {
            'label': 'Cavity_pole, f_cc (Hz)',
            'mathlabel': r'$f_{cc}$'},
        'springF': {
            'label': 'Detuned SRC spring frequency, f_s (Hz)',
            'mathlabel': r'$f_s$'},
        'springQ': {
            'label': 'Detuned SRC spring quality factor, Q_s',
            'mathlabel': r'$Q$'},
        'resDelay': {
            'label': 'Residual time delay, tau_c (s)',
            'mathlabel': r'$\Delta\tau_C$'},
    }

    # Set up the quantile levels
    quantileLevels = np.array([0.16, 0.5, 0.84])

    # Loop over params and store relevant information into the dict
    for i, param in enumerate(mcmcParams.values()):
        quantiles = corner.quantile(
            mcmcChains[i],
            quantileLevels)
        param['median'] = quantiles[1]
        param['errplus'] = quantiles[2]-quantiles[1]
        param['errminus'] = quantiles[1]-quantiles[0]
        param['quantiles'] = quantiles

    if mcmcParams['resDelay']['median'] >= 5e-3:  # TODO: make user arg
        logger.info(
            "Warning: MCMC fit for residual delay is high"
            f" at {mcmcParams['resDelay']['median']:.2f} s")

    # Make corner plot
    corner_fig = ru.make_corner_plot(
        mcmcChains, mcmcParams, quantileLevels,
        os.path.join(
            mset_dir,
            "sensing_mcmc_corner.png"),
        f"{mset_id} sensing function\nMCMC corner plot")

    # Add additional entry for kappa_c so that values can be printed
    mcmcParams['kappa_c'] = {
        'label': 'kappa_c',
        'mathlabel': r'$kappa_c$'
    }

    for key in ['median', 'errplus', 'errminus', 'quantiles']:
        mcmcParams['kappa_c'][key] = (
            mcmcParams['opticalGain'][key]
            / C_ref.coupled_cavity_optical_gain)

    # Print out and parameter value table
    mcmcTableQuant, mcmcTablePM = ru.print_mcmc_params(mcmcParams, quantileLevels)
    if config['mcmc_print']:
        logger.info(f"\n{mcmcTableQuant}")
        logger.info(f"\n{mcmcTablePM}\n")

    return mcmcParams, mcmcTableQuant, mcmcTablePM, corner_fig


def run_actuation_MCMC(
        processed_actuation, config, mset_id, stage, gain_units, gain_to_NpCt_factor, mset_dir):
    # TODO write proper doc string
    # Run the MCMC chain
    mcmcChains = np.transpose(
        processed_actuation.run_mcmc(
            fmin=config['mcmc_fmin'],
            fmax=config['mcmc_fmax'],
            burn_in_steps=config['mcmc_burn_in'],
            steps=config['mcmc_steps'],
            save_chain_to_file=os.path.join(
                mset_dir, f'actuation_{stage}_mcmc_chain.hdf5'),
            save_map_to_file=os.path.join(
                mset_dir, f'actuation_{stage}_mcmc_map.json'),
        ))

    # Set up a dict of dicts to hold the mcmc parameter information
    mcmcParams = {
        'gainNpDriveOutUnits': {
            'label': f'Gain, H_A ({gain_units})',
            'mathlabel': r'$H_A$'+f' ({gain_units})'},
        'resDelay': {
            'label': 'Residual time delay, tau_c (s)',
            'mathlabel': r'$\Delta\tau_C$'},
    }

    # Set up the quantile levels
    quantileLevels = np.array([0.16, 0.5, 0.84])

    # Loop over params and store relevant information into the dict
    for i, param in enumerate(mcmcParams.values()):
        quantiles = corner.quantile(
            mcmcChains[i],
            quantileLevels)
        param['median'] = quantiles[1]
        param['errplus'] = quantiles[2]-quantiles[1]
        param['errminus'] = quantiles[1]-quantiles[0]
        param['quantiles'] = quantiles

    if mcmcParams['resDelay']['median'] >= 5e-3:  # TODO: make user arg
        logger.info(
            "Warning: MCMC fit for residual delay is high"
            f" at {mcmcParams['resDelay']['median']:.2f} s")

    # Make corner plot
    corner_fname = f"actuation_{stage}_mcmc_corner.png"
    corner_title = f"{mset_id} {stage} actuation function\nMCMC corner plot"
    corner_fig = ru.make_corner_plot(mcmcChains,
                                     mcmcParams, quantileLevels,
                                     os.path.join(mset_dir, corner_fname),
                                     corner_title)

    # Add additional entry for gain in physical units so it can be printed
    mcmcParams['gainNpCt'] = {
        'label': 'Gain, H_A (N/ct)',
        'mathlabel': r'$H_A$ (N/ct)'
    }

    for key in ['median', 'errplus', 'errminus', 'quantiles']:
        mcmcParams['gainNpCt'][key] = (
            mcmcParams['gainNpDriveOutUnits'][key]
            * gain_to_NpCt_factor)

    # Print out and parameter value table
    mcmcTableQuant, mcmcTablePM = ru.print_mcmc_params(mcmcParams, quantileLevels)

    if config['mcmc_print']:
        logger.info(f"\n{mcmcTableQuant}")
        logger.info(f"\n{mcmcTablePM}\n")

    return mcmcParams, mcmcTableQuant, mcmcTablePM, corner_fig


def actuation(mset):
    """generate actuation function MCMC and plots

    """
    # TODO: write proper docstring
    logger.info("generating actuation report...")

    ifo = mset.model.name
    A_ref = mset.model.actuation

    # Format the reference model date and the filename
    # ref_tag = ru.extract_date_general(model_ini).strftime("%Y-%m-%d")
    ref_tag = ''
    subtitle_text = (
        f"All fixed parameters are drawn from"
        f" {os.path.basename(mset.model_ini)}")

    # Set up common title names
    tfp_title = "Actuation strength transfer functions"
    rp_title = "Actuation strength residuals"
    sp_titlesize = 12

    # figure list to return
    figs = []
    # process by actuator stage
    for stage in ['L1', 'L2', 'L3']:

        config = mset.config['Actuation'][stage]
        processed_actuation = mset.processed_measurements[f'Actuation/{stage}']

        # main figure
        plt.clf()
        # TODO: with multiple figs created here, choose better var names
        fig = plt.figure()
        fig.suptitle(
            f"{ifo}SUS{config['optic']}"
            f" {stage} actuation model history\n",
            fontsize=20)
        fig.text(
            .5, .93,
            subtitle_text,
            horizontalalignment='center',
            transform=fig.transFigure)

        # transfer function plot (comparison)
        # TODO: replace figure vars (tfp & rp) below with better names
        tfp = BodePlot(fig=fig, spspec=[221, 223])
        tfp.ax_mag.set_title(tfp_title, fontsize=sp_titlesize)

        # residuals plot (comparison)
        rp = BodePlot(fig=fig, spspec=[222, 224])
        rp.ax_mag.set_title(rp_title, fontsize=sp_titlesize)

        # In order to be generic and allow for both arms as actuators,
        # A has methods for xarm and yarm. But, if I'm only using one arm's
        # actuator, then I can reduce the complexity (and simplify later
        # code) by extract the layer of method I need and naming it generically
        if config['optic'] == 'ETMX':
            A_ref_arm = A_ref.xarm
        else:
            A_ref_arm = A_ref.yarm

        # gain_units, gain_to_NpCt_factor
        if stage == 'L1':
            ref_actuator_strength_param = abs(A_ref_arm.uim_npa)
            gain_to_NpCt_factor = abs(
                A_ref_arm.uim_dc_gain_Apct())
            gain_units = 'N/A'
        elif stage == 'L2':
            ref_actuator_strength_param = abs(A_ref_arm.pum_npa)
            gain_to_NpCt_factor = abs(
                A_ref_arm.pum_dc_gain_Apct())
            gain_units = 'N/A'
        else:
            ref_actuator_strength_param = abs(A_ref_arm.tst_npv2)
            gain_to_NpCt_factor = abs(
                A_ref_arm.tst_dc_gain_V2pct())
            gain_units = 'N/V**2'

        # Loop over measurement dates/times
        # for im, measXMLpair in enumerate(sensing_xml_pairs):
        if True:
            im = 0

            # Get the common frequency axis and measurement info
            frequencies, meas_actuator_strength, meas_actuator_strength_unc = \
                processed_actuation.get_processed_measurement_response()

            ref_actuator_strength = ref_actuator_strength_param * np.ones(
                frequencies.shape)

            if im == 0:

                # Select scaling for the plot
                expscale = int(np.floor(np.log10(ref_actuator_strength_param)))
                scale = 10**expscale
                tfp.ax_mag.set_ylabel(f'Magnitude x $10^{{{expscale}}}$')

                # Add reference model curve
                tfp.plot(frequencies, ref_actuator_strength/scale,
                         label=f"{ref_tag} model")

                # Add a null curve to keep the color-coding consistent
                # to the residuals plot
                rp.plot([], [])

            if (im == 0 and config['mcmc_mode'] in ['latest', 'all']):

                mcmcParams, mcmcTableQuant, mcmcTablePM, corner_fig = run_actuation_MCMC(
                    processed_actuation,
                    config,
                    mset.id,
                    stage,
                    gain_units,
                    gain_to_NpCt_factor,
                    mset.path,
                )
                figs.append(corner_fig)

                mcmc_actuator_strength = mcmcParams['gainNpDriveOutUnits']['median'] * np.exp(
                    -2*np.pi*1j*mcmcParams['resDelay']['median']*frequencies)

                fig_mcmc = plt.figure(figsize=(10, 10))
                fig_mcmc.suptitle(
                    f"{ifo}SUS{config['optic']} {stage} "
                    "actuation model MCMC summary\n",
                    fontsize=20)
                fig_mcmc.text(
                    .5, .93,
                    subtitle_text,
                    in_layout=True,
                    horizontalalignment='center',
                    transform=fig_mcmc.transFigure)

                tfp_mcmc = BodePlot(fig=fig_mcmc, spspec=[221, 223])
                tfp_mcmc.ax_mag.set_title(tfp_title, fontsize=sp_titlesize)
                tfp_mcmc.ax_mag.set_ylabel(f'Magnitude x $10^{{{expscale}}}$')
                rp_mcmc = BodePlot(fig=fig_mcmc, spspec=[222, 224])
                rp_mcmc.ax_mag.set_title(rp_title, fontsize=sp_titlesize)

                # Add the curves to the plot
                tfp_mcmc.plot(
                    frequencies,
                    ref_actuator_strength/scale,
                    label=(
                        "Model w free params from\n "
                        f"{os.path.basename(mset.model_ini)}"))
                tfp_mcmc.plot(
                    frequencies,
                    mcmc_actuator_strength/scale,
                    label=f"Model w free params from\n MCMC fit to {mset.id} data")
                tfp_mcmc.error(
                    frequencies,
                    meas_actuator_strength/scale,
                    meas_actuator_strength_unc,
                    label=f"{mset.id} measurement",
                    fmt='.')

                rp_mcmc.error(
                    frequencies,
                    meas_actuator_strength/mcmc_actuator_strength,
                    meas_actuator_strength_unc,
                    label=(
                        f"{mset.id} meas / model w free params\n"
                        f" from {os.path.basename(mset.model_ini)}"),
                    fmt='.')
                rp_mcmc.error(
                    frequencies,
                    meas_actuator_strength/ref_actuator_strength,
                    meas_actuator_strength_unc,
                    label=(
                        f"{mset.id} meas / model w free params\n"
                        f" from {os.path.basename(mset.model_ini)}"),
                    fmt='.')

                legends = []
                # Add vertical lines marking the fit range for the MCMC
                for p in [tfp_mcmc, rp_mcmc]:
                    p.vlines(
                        config['mcmc_fmin'], color='k', lw=2,
                        label=f"Fit range {config['mcmc_fmin']} to"
                        f" {config['mcmc_fmax']} Hz")
                    p.vlines(config['mcmc_fmax'], color='k', lw=2)
                    # Add legends
                    leg = p.ax_mag.legend(
                        bbox_to_anchor=(0, 1.1),
                        loc='lower left',
                        bbox_transform=p.ax_mag.transAxes)
                    legends += [leg]

                fig_mcmc.tight_layout()
                tbox = fig_mcmc.text(
                    .5, 0,
                    "\n"*6+mcmcTablePM+"\n",
                    fontfamily='monospace',
                    horizontalalignment='center',
                    verticalalignment='bottom',
                    transform=fig_mcmc.transFigure,
                )

                text_bbox = tbox.get_tightbbox()
                text_height = text_bbox.y1-text_bbox.y0
                fig_height = fig_mcmc.get_size_inches()[1]*fig_mcmc.dpi
                adjust_fraction = (text_height)/fig_height
                fig_mcmc.subplots_adjust(bottom=adjust_fraction)
                fig_mcmc.savefig(
                    os.path.join(
                        mset.path,
                        f"actuation_{stage}_mcmc_compare.png"),
                )
                figs.append(fig_mcmc)

            # Add meas curves to transfer function comparison plots
            tfp.error(
                frequencies,
                meas_actuator_strength/scale,
                meas_actuator_strength_unc,
                label=f"{mset.id} measurement",
                fmt='.')

            # Add meas curves to residuals plot
            rp.error(
                frequencies,
                meas_actuator_strength/ref_actuator_strength,
                meas_actuator_strength_unc,
                label=(
                    f"{mset.id} meas / model w free params\n"
                    f" from {os.path.basename(mset.model_ini)}"),
                fmt='.')

        # Add legends
        for p in [tfp, rp]:
            p.legend(
                bbox_to_anchor=(0, 1.1),
                loc='lower left',
                bbox_transform=p.ax_mag.transAxes)

        # Wrap up and save figure
        fig.tight_layout()
        fig.savefig(
            os.path.join(
                mset.path,
                f"actuation_{stage}_tf_history.png"))
        figs.append(fig)
    return figs


def sensing(mset):
    """generate sensing function MCMC and plots

    """
    # TODO: write proper docstring
    logger.info("generating sensing report...")

    ifo = mset.model.name
    config = mset.config['Sensing']
    processed_sensing = mset.processed_measurements['Sensing']

    # Format the reference model date and filename
    # refDateTag = ru.extract_date_general(model_ini).strftime("%Y-%m-%d")
    refDateTag = ''
    subtitleText = (f"All fixed parameters are drawn from"
                    f" {os.path.basename(mset.model_ini)}")

    # Set up common title names
    tfp_title = "Optical response transfer functions"
    rp_title = "Optical response residuals"
    sp_titlesize = 12

    # === Plotting setup

    # main figure
    plt.clf()
    fig = plt.figure()
    fig.suptitle(f"{ifo} sensing model history\n", fontsize=20)
    fig.text(
        .5, .93,
        subtitleText,
        horizontalalignment='center',
        transform=fig.transFigure)

    # transfer function plot (comparison)
    tfp = BodePlot(fig=fig, spspec=[221, 223])
    tfp.ax_mag.set_title(tfp_title, fontsize=sp_titlesize)

    # residuals plot (comparison)
    rp = BodePlot(fig=fig, spspec=[222, 224])
    rp.ax_mag.set_title(rp_title, fontsize=sp_titlesize)

    # figure list to return
    figs = [fig]

    # Create reference sensing model
    C_ref = SensingModel(mset.model_ini)

    # Get the common frequency axis and measurement info
    frequencies, measOpticalResponse, measOpticalResponseUnc = \
        processed_sensing.get_processed_measurement_response()

    angular_frequencies = 2*np.pi*frequencies

    # Compute the optical response based on reference parameters
    # Note: this reference optical response needs to be computed for
    # each measurement pair despite the fact that it is only plotted
    # once. This is because the frequency axis could differ between
    # measurements, and `refOpticalResponse` must divide the measurement
    # in order to generate the residuals plot.

    refNormOpticalResponse = freqresp(
        SensingModel.optical_response(
            C_ref.coupled_cavity_pole_frequency,
            C_ref.detuned_spring_frequency,
            C_ref.detuned_spring_q,
            C_ref.is_pro_spring),
        angular_frequencies)[1]

    refOpticalResponse = refNormOpticalResponse * \
        C_ref.coupled_cavity_optical_gain

    # If this is the first measurement, we need to compute the
    # reference model response. (This can't be done ahead of time
    # because it relies on the common frequency axis)
    # If requested, we will also run the MCMC on this first measurement.
    im = 0
    if im == 0:

        # Select scaling for the plot
        expscale = int(np.floor(np.log10(C_ref.coupled_cavity_optical_gain)))
        tfp.ax_mag.set_ylabel(f'Magnitude (ct/m) x $10^{{{expscale}}}$')
        scale = 10**expscale

        # Add reference model curve
        tfp.plot(frequencies, refOpticalResponse/scale, label=f"{refDateTag} model")

        # Add a null curve to keep the color-coding consistent on the residuals plot
        rp.plot([], [])

    # If requested, run the MCMC at this point
    fig_mcmc = None
    if (im == 0 and config['mcmc_mode'] == 'latest') \
       or config['mcmc_mode'] == 'all':

        # Run, return parameters, and create corner plot
        mcmcParams, mcmcTableQuant, mcmcTablePM, corner_fig = run_sensing_MCMC(
            processed_sensing,
            config,
            mset.id,
            C_ref,
            mset.path,
        )
        figs.append(corner_fig)

        # Compute the optical response based on MCMC parameters
        mcmcNormOpticalResponse = freqresp(
            SensingModel.optical_response(
                mcmcParams['cavityPole']['median'],
                mcmcParams['springF']['median'],
                mcmcParams['springQ']['median'],
                C_ref.is_pro_spring),
            angular_frequencies)[1]

        mcmcOpticalResponse = mcmcNormOpticalResponse * \
            mcmcParams['opticalGain']['median'] * \
            np.exp(
                -2*np.pi*1j*mcmcParams['resDelay']['median'] *
                frequencies)

        # We need an additional figure for the MCMC results comparison.
        # Setup is in the same format as the multi-measurement comparison.
        fig_mcmc = plt.figure(figsize=(10, 10))
        fig_mcmc.suptitle(f"{ifo} sensing model MCMC summary\n", fontsize=20)
        fig_mcmc.text(
            .5, .93,
            subtitleText,
            in_layout=True,
            horizontalalignment='center',
            transform=fig_mcmc.transFigure)

        tfp_mcmc = BodePlot(fig=fig_mcmc, spspec=[221, 223])
        tfp_mcmc.ax_mag.set_title(tfp_title, fontsize=sp_titlesize)
        tfp_mcmc.ax_mag.set_ylabel(f'Magnitude (ct/m) x $10^{{{expscale}}}$')
        rp_mcmc = BodePlot(fig=fig_mcmc, spspec=[222, 224])
        rp_mcmc.ax_mag.set_title(rp_title, fontsize=sp_titlesize)

        # Add the curves to the plot
        tfp_mcmc.plot(
            frequencies,
            refOpticalResponse/scale,
            label=f"Model w free params from\n {os.path.basename(mset.model_ini)}")
        tfp_mcmc.plot(
            frequencies,
            mcmcOpticalResponse/scale,
            label=f"Model w free params from\n MCMC fit to {mset.id} data")
        tfp_mcmc.error(
            frequencies,
            measOpticalResponse/scale,
            measOpticalResponseUnc,
            label=f"{mset.id} measurement / C_R",
            fmt='.')

        rp_mcmc.error(
            frequencies,
            measOpticalResponse/mcmcOpticalResponse,
            measOpticalResponseUnc,
            label=(
                f"{mset.id} meas / model w free params\n"
                f" from {os.path.basename(mset.model_ini)}"),
            fmt='.')
        rp_mcmc.error(
            frequencies,
            measOpticalResponse/refOpticalResponse,
            measOpticalResponseUnc,
            label=(
                f"{mset.id} meas / model w free params\n"
                f" from {os.path.basename(mset.model_ini)}"),
            fmt='.')

        legends = []
        # Add vertical lines marking the fit range for the MCMC
        for p in [tfp_mcmc, rp_mcmc]:
            p.vlines(
                config['mcmc_fmin'], color='k', lw=2,
                label=f"Fit range {config['mcmc_fmin']} to"
                f" {config['mcmc_fmax']} Hz")
            p.vlines(config['mcmc_fmax'], color='k', lw=2)
            # Add legends
            leg = p.ax_mag.legend(
                bbox_to_anchor=(0, 1.1),
                loc='lower left',
                bbox_transform=p.ax_mag.transAxes)
            legends += [leg]

        fig_mcmc.tight_layout()
        tbox = fig_mcmc.text(
            .5, 0,
            "\n"*6+mcmcTablePM+"\n",
            fontfamily='monospace',
            horizontalalignment='center',
            verticalalignment='bottom',
            transform=fig_mcmc.transFigure,
        )

        text_bbox = tbox.get_tightbbox()
        text_height = text_bbox.y1-text_bbox.y0
        fig_height = fig_mcmc.get_size_inches()[1]*fig_mcmc.dpi
        adjust_fraction = (text_height)/fig_height
        fig_mcmc.subplots_adjust(bottom=adjust_fraction)
        fig_mcmc.savefig(
            os.path.join(
                mset.path,
                "sensing_mcmc_compare.png"),
        )
        figs.append(fig_mcmc)

    # Add meas curves to transfer function comparison plots
    tfp.error(
        frequencies,
        measOpticalResponse/scale,
        measOpticalResponseUnc,
        label=f"{mset.id} measurement / C_R",
        fmt='.')

    # Add meas curves to residuals plot
    rp.error(
        frequencies,
        measOpticalResponse/refOpticalResponse,
        measOpticalResponseUnc,
        label=(
            f"{mset.id} meas / model w free params\n"
            f" from {os.path.basename(mset.model_ini)}"),
        fmt='.')

    # Add legends
    for p in [tfp, rp]:
        p.legend(
            bbox_to_anchor=(0, 1.1),
            loc='lower left',
            bbox_transform=p.ax_mag.transAxes)

    # Wrap up and save figure
    fig.tight_layout()
    fig.savefig(
        os.path.join(
            mset.path,
            "sensing_tf_history.png"))
    return figs


class ReportPDF:

    def __init__(self, path):
        self.path = path

    def __enter__(self):
        self.pdf = PdfPages(self.path)
        return self

    def add_figure(self, new_fig):
        if hasattr(new_fig, '__iter__'):
            for f in new_fig:
                self.pdf.savefig(f)
        else:
            self.pdf.savefig(new_fig)

    def __exit__(self, exception_type, exception_value, traceback):
        self.pdf.close()
        # remove the report if there was an exception, so that we
        # don't leave around a malformed report.
        # FIXME: confirm this is what we want
        if exception_type is not None:
            os.unlink(self.path)
        return False


def add_args(parser):
    parser.add_argument(
        'mset_id', metavar='MSET_ID', nargs='?',
        help="measurement set ID to view/process, or last measurement set if not specified")
    _util.add_model_option(parser)
    _util.add_config_option(parser)
    parser.add_argument(
        '--force', action='store_true',
        help="force generate report if it already exists")
    parser.add_argument(
        '--display', action=argparse.BooleanOptionalAction, default=True,
        help="display report after generation or not (display by default)")


def main(args):
    """generate full calibration report

    """
    mset = MeasurementSet(args.mset_id, args.config)
    logger.info(f"measurement set: {mset.id}")

    report_file = os.path.join(
        mset.path,
        f"{IFO}_calibration_report_{args.mset_id}.pdf",
    )

    if not os.path.exists(report_file) or args.force:
        if os.path.exists(report_file):
            value = input(f"Really regenerate report for measurement set {mset.id} [Y|n]? ")
            if value.upper() not in ['Y', '']:
                logger.info("aborting.")
                return
        logger.info(f"generating report for {mset.id}...")
        mset.load_model(args.model)
        with ReportPDF(report_file) as report:
            report.add_figure(sensing(mset))
            report.add_figure(actuation(mset))
        logger.info("report generation complete.")

    if args.display:
        logger.info(f"displaying report {report_file}")
        # FIXME: make viewer this configurable
        subprocess.run(['evince', '-w', report_file])



import pydarm
import numpy as np
from gwpy.timeseries import TimeSeriesDict as tsd
import os

os.environ['CAL_DATA_ROOT'] = '../test'

frequencies = np.logspace(0, np.log10(5000.), 10)

# Reading in data from GWPY
# This data was previously saved, and is completely optional to provide
data = tsd.read('../test/test_ch_data.hdf5')

# See ../example_model_files/H1_20190416_uncertainty.ini
# for more details
config = '''[reference-model]
model = ../example_model_files/H1_20190416.ini
[sensing-measurement-files]
mcmc =
gpr =
[x-arm-measurement-files]
tst_mcmc =
pum_mcmc =
uim_mcmc =
tst_gpr =
pum_gpr =
uim_gpr =
[x-arm-sus-cal-lines]
uim = 15.6
pum = 16.4
tst = 17.6
[y-arm-pcal-cal-lines]
sus = 17.1
sensing = 410.3
pcal3 = 1083.7
pcal4 = 7.93
[tdcf-data]
frametype = R
duration = 130
[sensing-tdcf]
kappa_c = CAL-CS_TDEP_KAPPA_C_OUTPUT
f_cc = CAL-CS_TDEP_F_C_OUTPUT
pcal2_unc = CAL-CS_TDEP_PCAL_LINE2_UNCERTAINTY
pcal_arm = Y
[x-arm-tdcf]
kappa_uim = CAL-CS_TDEP_KAPPA_UIM_REAL_OUTPUT
kappa_pum = CAL-CS_TDEP_KAPPA_PUM_REAL_OUTPUT
kappa_tst = CAL-CS_TDEP_KAPPA_TST_REAL_OUTPUT
pcal1_unc = CAL-CS_TDEP_PCAL_LINE1_UNCERTAINTY
uim_unc = CAL-CS_TDEP_SUS_LINE1_UNCERTAINTY
pum_unc = CAL-CS_TDEP_SUS_LINE2_UNCERTAINTY
tst_unc = CAL-CS_TDEP_SUS_LINE3_UNCERTAINTY
pcal_arm = Y
[sample-tdcf]
kappa_c = False
f_cc = False
kappa_uim = False
kappa_pum = False
kappa_tst = False
[hoft-tdcf-data-application]
kappa_c = True
f_cc = True
kappa_uim = True
kappa_pum = True
kappa_tst = True
pcal_sys_err = False
[pcal]
sys_err =
sys_unc = 0.005
sample = True'''

# Intialize a DARMUncertainy object
test_unc = pydarm.uncertainty.DARMUncertainty(config)

# Sample the uncertainty
samples = test_unc.compute_response_uncertainty(
    1239958818, frequencies, trials=10, data=data)



import unittest
import os
import numpy as np
from scipy import signal
from pydarm.utils import (serielZPK, parallelZPK,
                          freqrespZPK, dfreqrespZPK,
                          digital_delay_filter,
                          save_chain_to_hdf5,
                          read_chain_from_hdf5,
                          save_gpr_to_hdf5,
                          read_gpr_from_hdf5,
                          read_eta_or_syserr_from_hdf5,
                          thiran_delay_filter)


class TestSerielZPK(unittest.TestCase):

    def setUp(self):
        self.filter1 = signal.ZerosPolesGain(
            -2.0*np.pi*np.array([1.0, 1.0]),
            -2.0*np.pi*np.array([100.0, 100.0]),
            100.0**2/1.0**2)
        self.filter2 = signal.ZerosPolesGain(
            -2.0*np.pi*np.array([2.0, 2.0]),
            -2.0*np.pi*np.array([200.0, 200.0]),
            200.0**2/2.0**2)
        self.filter = signal.ZerosPolesGain(
            -2.0*np.pi*np.array([1.0, 1.0, 2.0, 2.0]),
            -2.0*np.pi*np.array([100.0, 100.0, 200.0, 200.0]),
            (100.0**2*200.0**2/(1.0**2*2.**2)))

    def tearDown(self):
        del self.filter1
        del self.filter2
        del self.filter

    def test_serielZPK(self):
        test_filter = serielZPK(self.filter1, self.filter2)
        self.assertEqual(len(test_filter.zeros), len(self.filter.zeros))
        self.assertEqual(len(test_filter.poles), len(self.filter.poles))
        for n in range(len(self.filter.zeros)):
            self.assertAlmostEqual(test_filter.zeros[n], self.filter.zeros[n])
        for n in range(len(self.filter.poles)):
            self.assertAlmostEqual(test_filter.poles[n], self.filter.poles[n])
        self.assertAlmostEqual(test_filter.gain, self.filter.gain)


class TestParallelZPK(unittest.TestCase):

    def setUp(self):
        self.filter1 = signal.ZerosPolesGain(
            -2.0*np.pi*np.array([1.0, 1.0]),
            -2.0*np.pi*np.array([100.0, 100.0]),
            100.0**2/1.0**2)
        self.filter2 = signal.ZerosPolesGain(
            -2.0*np.pi*np.array([2.0, 2.0]),
            -2.0*np.pi*np.array([200.0, 200.0]),
            200.0**2/2.0**2)
        self.filter = signal.ZerosPolesGain(
            np.array([-944.3717956376961+313.51778225283215j,
                      -944.3717956376961-313.51778225283215j,
                      -7.530778400011287+2.5001095474423356j,
                      -7.530778400011287-2.5001095474423356j]),
            np.array([-1256.6370614359168+5.624423698413676e-05j,
                      -1256.6370614359168-5.624423698413676e-05j,
                      -628.3185344963764+0j,
                      -628.3185269395391+0j]),
            20000.0)

    def tearDown(self):
        del self.filter1
        del self.filter2
        del self.filter

    def test_parallelZPK(self):
        test_filter = parallelZPK(self.filter1, self.filter2)
        self.assertEqual(len(test_filter.zeros), len(self.filter.zeros))
        self.assertEqual(len(test_filter.poles), len(self.filter.poles))
        for n in range(len(self.filter.zeros)):
            self.assertAlmostEqual(np.real(test_filter.zeros[n]),
                                   np.real(self.filter.zeros[n]))
            self.assertAlmostEqual(np.imag(test_filter.zeros[n]),
                                   np.imag(self.filter.zeros[n]))
        for n in range(len(self.filter.poles)):
            self.assertAlmostEqual(np.real(test_filter.poles[n]),
                                   np.real(self.filter.poles[n]), places=3)
            self.assertAlmostEqual(np.imag(test_filter.poles[n]),
                                   np.imag(self.filter.poles[n]), places=3)
        self.assertAlmostEqual(test_filter.gain, self.filter.gain)


class TestFreqrespZPK(unittest.TestCase):

    def setUp(self):
        self.filter = signal.ZerosPolesGain(
            [], -2.0*np.pi*np.array([1, 1]),
            np.prod(2.0*np.pi*np.asarray([1, 1])))
        self.frequencies = np.logspace(0, np.log10(5000.), 10)
        self.known_response = np.array(
            [0.0-0.5j,
             -0.09664744749943084-0.08833721090114179j,
             -0.021210239147445683-0.0065396429418788145j,
             -0.003385062727101369-0.00039727801004293907j,
             -0.0005144644812257184-2.336805444122701e-05j,
             -7.761259172380499e-05-1.367768387548693e-06j,
             -1.1695660566624151e-05-7.999812866147877e-08j,
             -1.762156044886542e-06-4.678411892855308e-09j,
             -2.654929417882733e-07-2.7359600619251977e-10j,
             -3.999999520000035e-08-1.5999998720000092e-11j])

    def tearDown(self):
        del self.filter
        del self.frequencies
        del self.known_response

    def test_freqrespZPK(self):
        test_response = freqrespZPK(self.filter, 2.0*np.pi*self.frequencies)
        for n in range(len(test_response)):
            self.assertAlmostEqual(np.abs(test_response[n]),
                                   np.abs(self.known_response[n]))
            self.assertAlmostEqual(np.angle(test_response[n], deg=True),
                                   np.angle(self.known_response[n], deg=True))


class TestDfreqrespZPK(unittest.TestCase):

    def setUp(self):
        self.filter = signal.ZerosPolesGain(
            [], np.zeros(4), 1, dt=1.0/2**16)
        self.frequencies = np.logspace(0, np.log10(5000.), 10)
        self.known_response = np.array(
            [0.9999999264657179-0.00038349518757139556j,
             0.9999995119288227-0.0009879990467211066j,
             0.9999967605125888-0.002545380978905176j,
             0.9999784985222522-0.006557628624898654j,
             0.9998572905048693-0.01689374512241721j,
             0.9990529171554058-0.04351170788734579j,
             0.9937194923983442-0.11189982317938384j,
             0.9585599169648092-0.2848910065067314j,
             0.7355794304443074-0.6774384854045631j,
             -0.3397768844068262-0.9405060705932685j])

    def tearDown(self):
        del self.filter
        del self.frequencies
        del self.known_response

    def test_dfreqrespZPK(self):
        test_response = dfreqrespZPK(self.filter,
                                     2.0*np.pi*self.frequencies/2**16)[1]
        for n in range(len(test_response)):
            self.assertAlmostEqual(np.abs(test_response[n]),
                                   np.abs(self.known_response[n]))
            self.assertAlmostEqual(np.angle(test_response[n], deg=True),
                                   np.angle(self.known_response[n], deg=True))


class TestDigitalDelayFilter(unittest.TestCase):

    def setUp(self):
        self.delay_cycles = 4
        self.filter = signal.ZerosPolesGain(
            [], np.zeros(self.delay_cycles), 1, dt=1.0/2**16)

    def tearDown(self):
        del self.delay_cycles
        del self.filter

    def test_digial_delay_filter(self):
        test_filter = digital_delay_filter(self.delay_cycles, 2**16)
        self.assertEqual(len(test_filter.zeros), len(self.filter.zeros))
        self.assertEqual(len(test_filter.poles), len(self.filter.poles))
        for n in range(len(test_filter.poles)):
            self.assertAlmostEqual(test_filter.poles[n], self.filter.poles[n])


class TestSaveChainToHdf5(unittest.TestCase):

    def setUp(self):
        self.filename = 'test.h5'
        self.model_string = 'test'
        self.fmin = 20
        self.fmax = 5000
        self.measurement = 'sensing'
        self.chain = np.arange(10)

    def tearDown(self):
        del self.filename
        del self.model_string
        del self.fmin
        del self.fmax
        del self.measurement
        del self.chain

    def test_save_chain_to_hdf5(self):
        save_chain_to_hdf5(
            self.filename, self.model_string, self.fmin, self.fmax,
            self.measurement, self.chain)

        assert os.path.exists(self.filename)


class TestReadChainFromHdf5(unittest.TestCase):

    def setUp(self):
        self.filename = 'test.h5'
        self.model_string = 'test'
        self.fmin = 20
        self.fmax = 5000
        self.measurement = 'sensing'
        self.chain = np.arange(10)

    def tearDown(self):
        del self.filename
        del self.model_string
        del self.fmin
        del self.fmax
        del self.measurement
        del self.chain

    def test_save_chain_to_hdf5(self):
        save_chain_to_hdf5(
            self.filename, self.model_string, self.fmin, self.fmax,
            self.measurement, self.chain)

        test = read_chain_from_hdf5(self.filename)

        assert test[0] == self.model_string
        assert test[1] == self.fmin
        assert test[2] == self.fmax
        assert test[3] == self.measurement
        assert np.allclose(test[4], self.chain)


class TestSaveGPRToHDF5(unittest.TestCase):

    def setUp(self):
        self.filename = 'test_GPR.h5'
        self.meas_model = 'test'
        self.measurement = 'sensing'
        self.known_y_pred = np.array(
             [1.1275231 -0.14775j, 1.11947795-0.1391194j])
        self.known_cov = np.array(
             [[ 1.20202662e-03, 1.11412775e-03, 1.06308187e-03, 1.01441776e-03,
                9.46456817e-04, 9.01659923e-04, 8.68963687e-04, 8.27314896e-04,
                7.66540580e-04, 7.29937311e-04, 6.03345479e-04, 5.34071642e-04,
                4.88979305e-04, 4.51309020e-04, 2.66398796e-04, 2.33383085e-04,
                2.07102041e-04, 1.89673520e-04, 1.43273106e-04, 4.98941061e-05,
                1.10710651e-05, -5.57882617e-05, -8.43420325e-05, -9.91110312e-05,
               -1.02160096e-04, -9.56746538e-05, -8.18911186e-05, -6.66049353e-05,
               -5.83790161e-05, -4.12302913e-05, -2.70112688e-05, -8.50187956e-06,
                3.36873466e-06, 2.33652634e-05, 3.84648854e-05, 5.48991959e-05,
                5.52764965e-05, 5.13437276e-05, 5.04097799e-05, 4.82236629e-05,
                4.76158458e-05, 1.78903912e-05, -2.83320392e-05, -6.83616761e-05],
              [ 1.11412775e-03, 1.03573118e-03, 9.90122350e-04, 9.46583654e-04,
                8.85681902e-04, 8.45472537e-04, 8.16090363e-04, 7.78619630e-04,
                7.23850749e-04, 6.90809621e-04, 5.76190547e-04, 5.13215977e-04,
                4.72115876e-04, 4.37709440e-04, 2.67681956e-04, 2.37077016e-04,
                2.12647158e-04, 1.96409834e-04, 1.53020146e-04, 6.47648751e-05,
                2.75239570e-05, -3.81327643e-05, -6.76286278e-05, -8.44823971e-05,
               -9.04436045e-05, -8.73904016e-05, -7.72683371e-05, -6.49851674e-05,
               -5.81547009e-05, -4.35844053e-05, -3.12326276e-05, -1.48324160e-05,
               -4.12176291e-06, 1.43260424e-05, 2.87599399e-05, 4.63037980e-05,
                4.77436656e-05, 4.58545026e-05, 4.52409985e-05, 4.37349213e-05,
                4.33033826e-05, 1.99378176e-05, -1.90976763e-05, -5.40446567e-05]])
        self.known_frequencies = np.array([10, 20])

    def tearDown(self):
        del self.filename
        del self.meas_model 
        del self.measurement
        del self.known_y_pred
        del self.known_cov
        del self.known_frequencies

    def test_save_gpr_to_hdf5(self):
        save_gpr_to_hdf5(self.filename,
                         self.meas_model,
                         self.measurement,
                         self.known_y_pred,
                         self.known_cov,
                         self.known_frequencies)

        assert os.path.exists(self.filename)


class TestReadGPRFromHDF5(unittest.TestCase):

    def setUp(self):
        self.filename = 'test_GPR.h5'
        self.meas_model = 'test'
        self.measurement = 'sensing'
        self.known_y_pred = np.array(
             [1.1275231 -0.14775j, 1.11947795-0.1391194j])
        self.known_cov = np.array(
             [[ 1.20202662e-03, 1.11412775e-03, 1.06308187e-03, 1.01441776e-03,
                9.46456817e-04, 9.01659923e-04, 8.68963687e-04, 8.27314896e-04,
                7.66540580e-04, 7.29937311e-04, 6.03345479e-04, 5.34071642e-04,
                4.88979305e-04, 4.51309020e-04, 2.66398796e-04, 2.33383085e-04,
                2.07102041e-04, 1.89673520e-04, 1.43273106e-04, 4.98941061e-05,
                1.10710651e-05, -5.57882617e-05, -8.43420325e-05, -9.91110312e-05,
               -1.02160096e-04, -9.56746538e-05, -8.18911186e-05, -6.66049353e-05,
               -5.83790161e-05, -4.12302913e-05, -2.70112688e-05, -8.50187956e-06,
                3.36873466e-06, 2.33652634e-05, 3.84648854e-05, 5.48991959e-05,
                5.52764965e-05, 5.13437276e-05, 5.04097799e-05, 4.82236629e-05,
                4.76158458e-05, 1.78903912e-05, -2.83320392e-05, -6.83616761e-05],
              [ 1.11412775e-03, 1.03573118e-03, 9.90122350e-04, 9.46583654e-04,
                8.85681902e-04, 8.45472537e-04, 8.16090363e-04, 7.78619630e-04,
                7.23850749e-04, 6.90809621e-04, 5.76190547e-04, 5.13215977e-04,
                4.72115876e-04, 4.37709440e-04, 2.67681956e-04, 2.37077016e-04,
                2.12647158e-04, 1.96409834e-04, 1.53020146e-04, 6.47648751e-05,
                2.75239570e-05, -3.81327643e-05, -6.76286278e-05, -8.44823971e-05,
               -9.04436045e-05, -8.73904016e-05, -7.72683371e-05, -6.49851674e-05,
               -5.81547009e-05, -4.35844053e-05, -3.12326276e-05, -1.48324160e-05,
               -4.12176291e-06, 1.43260424e-05, 2.87599399e-05, 4.63037980e-05,
                4.77436656e-05, 4.58545026e-05, 4.52409985e-05, 4.37349213e-05,
                4.33033826e-05, 1.99378176e-05, -1.90976763e-05, -5.40446567e-05]])
        self.known_frequencies = np.array([10, 20])

    def tearDown(self):
        del self.filename
        del self.meas_model
        del self.measurement
        del self.known_y_pred
        del self.known_cov
        del self.known_frequencies

    def test_read_gpr_from_hdf5(self):
        save_gpr_to_hdf5(self.filename,
                         self.meas_model,
                         self.measurement,
                         self.known_y_pred,
                         self.known_cov,
                         self.known_frequencies)

        test = read_gpr_from_hdf5(self.filename, self.measurement)

        self.assertTrue(test[0] == self.meas_model)
        for n in range(len(test[1])):
            self.assertAlmostEqual(
                np.abs(test[1][n])/np.abs(self.known_y_pred[n]), 1.0, places=4)
            self.assertAlmostEqual(np.angle(test[1][n], deg=True) -
                np.angle(self.known_y_pred[n], deg=True), 0.0, places=4)
        for n in range(len(test[2])):
            self.assertTrue(np.allclose(test[2][n], self.known_cov[n]))
        self.assertTrue(np.allclose(test[3], self.known_frequencies))


class TestReadEtaOrSysErrFromHDF5(unittest.TestCase):

    def setUp(self):
        self.filename = \
            '20220405_H1_SUSETMX_UIMCoilDriver_S0900303_40Ohm_State1_etaUIM_rdxmodelwsw.hdf5'
        self.measurement = 'etaUIM_rdxmodelwsw'
        self.known_frequencies = np.array(
            [2.000000e-01, 2.111752e-01,
             2.229748e-01, 2.354338e-01,
             2.485889e-01, 2.624790e-01,
             2.771453e-01, 2.926311e-01,
             3.089821e-01, 3.262468e-01])
        self.known_eta = np.array(
            [0.99990516+2.95568608e-04j, 0.99993306+2.78773978e-04j,
             0.99992513+2.62781436e-04j, 0.99990313+2.45549202e-04j,
             0.99988401+2.31071295e-04j, 0.99989508+2.17964399e-04j,
             0.9998946+2.06521167e-04j, 0.99990324+1.93850296e-04j,
             0.99987806+1.82878693e-04j, 0.99988041+1.75810276e-04j])

    def tearDown(self):
        del self.filename
        del self.measurement
        del self.known_frequencies
        del self.known_eta

    def test_read_eta_or_syserr_from_hdf5(self):

        test_frequencies, test_eta = \
            read_eta_or_syserr_from_hdf5('./test/'+self.filename, self.measurement)

        for n in range(len(self.known_frequencies)):
            self.assertAlmostEqual(
                np.abs(test_eta[n])/np.abs(self.known_eta[n]), 1.0, places=4)
            self.assertAlmostEqual(np.angle(test_eta[n], deg=True) -
                np.angle(self.known_eta[n], deg=True), 0.0, places=4)


class TestThiranDelayFilter(unittest.TestCase):

    def setUp(self):
        self.tau = 2.5
        self.Ts = 0.5
        self.filter = signal.TransferFunction([0, 0, 0, 0, 0, 1], [1, 0, 0, 0, 0, 0], dt=self.Ts)

    def tearDown(self):
        del self.tau
        del self.Ts
        del self.filter

    def test_thiran_delay_filter(self):
        test_filter = thiran_delay_filter(self.tau, self.Ts)
        self.assertEqual(len(test_filter.num), len(self.filter.num))
        self.assertEqual(len(test_filter.den), len(self.filter.den))
        self.assertEqual(test_filter.dt, self.filter.dt)
        for n in range(len(test_filter.num)):
            self.assertAlmostEqual(test_filter.num[n], self.filter.num[n])
            self.assertAlmostEqual(test_filter.den[n], self.filter.den[n])


if __name__ == '__main__':
    unittest.main()



import os
import shutil
import subprocess

from ._const import (
    CAL_TEMPLATE_ROOT,
    CAL_MEASUREMENT_ROOT,
)
from . import _util
from ._util import logger
from ._util import CMDError
from .measurement_set import template_str_replace


def diag_run(run_xml, save_xml=None, dry=False):
    """execute a DTT measurement template

    """
    if not save_xml:
        save_xml = run_xml
    diag_cmd = f'''
open
restore {run_xml}
run -w
save {save_xml}
quit
'''
    print(diag_cmd)
    if not dry:
        subprocess.run(
            ['diag'],
            input=diag_cmd.strip(),
            text=True,
        )
        print()
    # FIXME: check for errors somehow


def touch(path):
    """touch a file"""
    open(path, 'w').close()

##################################################


def add_args(parser):
    _util.add_config_option(parser)
    parser.add_argument(
        '--measure', '-m', nargs='*',
        help="measurements to be run (by default user default sequence from config)")
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        '--dry-run', '--dry', action='store_const', dest='run', const='dry',
        help="run measurement in headless mode")
    group.add_argument(
        '--run-interactive', action='store_const', dest='run', const='interactive',
        help="run measurement in interactive mode")
    group.add_argument(
        '--run-headless', action='store_const', dest='run', const='headless',
        help="run measurement in headless mode")


def main(args):
    """run calibration plant transer function measurements

    This command by default runs the full set of sensing and actuation
    chain calibration measurements.  The measurement config and
    templates are retrieved from the CAL_CONFIG_ROOT directory.  Each
    time a measurement is initiated a timestamped directory is created
    in the CAL_DATA_ROOT/measurements/ directory, and all resulting
    measurement output files are saved to that directory.

    When run without arguments the available measurements, the
    templates they use, and the default measurement sequence will be
    printed.

    A '--run-*' argument must be provided to actually run the
    measurements.

    """
    # https://cdswiki.ligo-wa.caltech.edu/wiki/TakingCalibrationMeasurements

    if not CAL_TEMPLATE_ROOT:
        raise CMDError("CAL_TEMPLATE_ROOT not specified.")
    if not CAL_MEASUREMENT_ROOT:
        raise CMDError("CAL_MEASUREMENT_ROOT not specified.")

    config = _util.load_config(args.config)

    measurements = {}
    for nick, mref in config['Common']['measurements'].items():
        section = config
        for name in mref.split('/'):
            section = section[name]
        measurements[nick] = {
            'description': section['description'],
            'template': os.path.join(
                CAL_TEMPLATE_ROOT,
                section['template'],
            ),
        }

    logger.info("available measurements:")
    for meas, conf in measurements.items():
        print(f"  {meas:<4}: {conf['description']} [{conf['template']}]")
        # check that all template files exist
        if not os.path.exists(conf['template']):
            raise CMDError(f"measurement '{meas}' template file not found: {conf['template']}")

    # construct the measurement sequence
    default_measurement_sequence = config['Common']['measurement_sequence']

    if not args.measure:
        mseq = default_measurement_sequence
    else:
        mseq = args.measure

    for meas in mseq:
        if meas not in measurements:
            raise CMDError(f"'{meas}' is not a valid measurement.")

    logger.info("measurement sequence:")
    logger.info(f"  {mseq}")

    if not args.run:
        logger.warning("Use --run-* option to actually execute measurement.")
        return

    ##########

    if args.run == 'dry':
        logger.warning("MEASUREMENT DRY RUN, use --run option to actually execute measurement")

    # ISO timestamp as measurement ID
    mset_id = _util.gen_timestamp()
    logger.info(f"measurement ID: {mset_id}")

    output_dir = os.path.join(CAL_MEASUREMENT_ROOT, mset_id)

    if args.run not in [None, 'dry']:
        if os.path.exists(output_dir):
            raise CMDError(f"output directory already exists: {output_dir}")
        os.makedirs(output_dir, exist_ok=False)
    logger.info(f"output directory: {output_dir}")

    # FIXME: check/verify IFO is in correct state for measurement:
    # e.g. ISC_LOCK == NLN_CAL_MEAS

    # FIXME: warning for thermalization status (less than an hour in
    # NLN)

    # FIXME: print expected measurement time
    # FIXME: progress bar?

    # FIXME: capture IFO config (EPICS/foton/etc) at measurement time

    ##########
    # loop over measurements
    for meas in mseq:
        logger.info("##########")

        measurement = measurements[meas]
        description = measurement['description']
        template_path = measurement['template']
        meas_timestamp = _util.gen_timestamp()
        output_path = os.path.join(
            output_dir,
            template_str_replace(os.path.basename(template_path), meas_timestamp),
        )

        logger.info(f"measurement: {meas}: {description} measurement")
        logger.info(f"measurement template: {template_path}")
        logger.info(f"measurement timestamp: {meas_timestamp}")
        logger.info(f"measurement output: {output_path}")

        if args.run not in [None, 'dry']:
            logger.info(f"executing {meas} measurement {args.run}...")

            # interactive running
            if args.run == 'interactive':
                tmp_path = output_path + '.tmp'
                shutil.copy(template_path, tmp_path)
                logger.warning("executing diaggui for interactive usage")
                logger.warning("save file to SAME PATH and exit to continue...")
                cmd = ['diaggui', tmp_path]
                logger.debug(cmd)
                subprocess.run(
                    cmd,
                )
                os.rename(tmp_path, output_path)

            # headless run
            elif args.run == 'headless':
                diag_run(template_path, output_path, dry=not args.run)

            # FIXME: somehow capture measurement errors

            logger.info(f"{meas} measurement complete.")
        else:
            continue

        # FIXME: validate the output somehow
        # FIXME: should we continue with the rest of the measurments
        # if there was an issue?

        if not os.path.exists(output_path):
            raise CMDError(f"output measurement file not found: {output_path}")
        logger.info(f"{meas} output: {output_path}")

    if args.run in [None, 'dry']:
        return

    # touch a file to indicate that all measurements were completed
    # successfully
    touch(os.path.join(output_dir, 'completed'))

    # touch a file to indicate that this measurement set corresponds
    # to a complete cal measurement set
    if mseq == default_measurement_sequence:
        touch(os.path.join(output_dir, 'full_measurement'))

    logger.info(f"all measurements complete: {output_dir}")



# Copyright (C) Evan Goetz (2021)
#
# This file is part of pyDARM.
#
# pyDARM is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pyDARM is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pyDARM. If not, see <https://www.gnu.org/licenses/>.

from . import (
    sensing,
    actuation,
    utils,
    darm,
    pcal,
    measurement,
    uncertainty,
    plot,
    calcs,
    FIR,
    FIRtools,
)


try:
    from ._version import version as __version__
except ModuleNotFoundError:
    try:
        import setuptools_scm
        __version__ = setuptools_scm.get_version(fallback_version='?.?.?')
    except (ModuleNotFoundError, TypeError, LookupError):
        __version__ = '?.?.?'

__author__ = 'Evan Goetz <evan.goetz@ligo.org>'
__credits__ = 'Ethan Payne, Antonios Kontos, Jameson Rollins, ' \
  'Miftahul Maarif, Afif Ismail, Hsiang-Yu Huang'



# Copyright (C) Evan Goetz (2021)
#
# This file is part of pyDARM.
#
# pyDARM is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# pyDARM is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# pyDARM. If not, see <https://www.gnu.org/licenses/>.

import numpy as np

from scipy import signal, io

from .utils import digital_delay_filter
from .utils import compute_digital_filter_response, freqrespZPK
from .model import Model


class ActuationModel(Model):
    """
    An arm actuation model object

    This is a class to set up the model for the actuation function from a
    configuration file with all the information about where the data is stored

    """

    def __init__(self, config, measurement):
        super().__init__(config, measurement=measurement)

        # TODO need some way of validating the configuration file is
        # appropriate

        # Setting up to the list to save loading in the suspension filters
        self.tst_digital_filter_response_list = []
        self.pum_digital_filter_response_list = []
        self.uim_digital_filter_response_list = []
        self.filter_response_freq_list = []

    def analog_driver_response(self, frequencies):
        """
        The transfer function of the analog driver electronics. By default,
        the output will be a dictionary for UIM, PUM, and TST with
        dictionaries for each containing the quadrants UL, LL, UR, LR.
        Transfer functions are unity with zero phase unless the params
        file/string contains a <uim|pum|tst>_driver_meas_<z|p>_<ul|ll|ur|lr>
        for the particular quadrant and values

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        out : dict, `complex128`, array-like
            transfer function response of the uncompensated driver electronics

        """

        out = {'UIM': {'UL': np.ones(len(frequencies), dtype='complex128'),
                       'LL': np.ones(len(frequencies), dtype='complex128'),
                       'UR': np.ones(len(frequencies), dtype='complex128'),
                       'LR': np.ones(len(frequencies), dtype='complex128')},
               'PUM': {'UL': np.ones(len(frequencies), dtype='complex128'),
                       'LL': np.ones(len(frequencies), dtype='complex128'),
                       'UR': np.ones(len(frequencies), dtype='complex128'),
                       'LR': np.ones(len(frequencies), dtype='complex128')},
               'TST': {'UL': np.ones(len(frequencies), dtype='complex128'),
                       'LL': np.ones(len(frequencies), dtype='complex128'),
                       'UR': np.ones(len(frequencies), dtype='complex128'),
                       'LR': np.ones(len(frequencies), dtype='complex128')}}

        for i, stage in enumerate(out.keys()):
            for j, quadrant in enumerate(out[stage].keys()):

                # Emit error if trying to use the old format
                if ((hasattr(self, f'{stage.lower()}_driver_meas_z_'
                                   f'{quadrant.lower()}') and
                     not hasattr(self, f'{stage.lower()}_driver_uncompensated_'
                                       f'z_{quadrant.lower()}')) or
                    (hasattr(self, f'{stage.lower()}_driver_meas_p_'
                                   f'{quadrant.lower()}') and
                     not hasattr(self, f'{stage.lower()}_driver_uncompensated_'
                                       f'p_{quadrant.lower()}'))):
                    raise KeyError(
                        'Using old format for zeros and poles. Please check'
                        ' your configuration string/file and use the'
                        ' updated format to specify'
                        ' compensated/uncompensated parameters. '
                        f'Stage: {stage.lower()}, Quadrant: {quadrant.lower()}')

                # Start with empty lists for zeros and poles. Extend as needed
                zeros = []
                poles = []
                if (hasattr(self, f'{stage.lower()}_driver_uncompensated_z_'
                                  f'{quadrant.lower()}') and
                        hasattr(self, f'{stage.lower()}_driver_uncompensated_'
                                      f'p_{quadrant.lower()}')):
                    zeros.extend(getattr(
                        self, f'{stage.lower()}_driver_uncompensated_z_'
                              f'{quadrant.lower()}'))
                    poles.extend(getattr(
                        self, f'{stage.lower()}_driver_uncompensated_p_'
                              f'{quadrant.lower()}'))
                if (hasattr(self, f'{stage.lower()}_driver_compensated_z_'
                                  f'{quadrant.lower()}') and
                        getattr(self, f'{stage.lower()}_front_end_driver_'
                                      f'compensation_{quadrant.lower()}') == 'OFF'):
                    zeros.extend(getattr(self, f'{stage.lower()}_driver_'
                                               f'compensated_z_{quadrant.lower()}'))
                if (hasattr(self, f'{stage.lower()}_driver_compensated_p_'
                                  f'{quadrant.lower()}') and
                        getattr(self, f'{stage.lower()}_front_end_driver_'
                                      f'compensation_{quadrant.lower()}') == 'OFF'):
                    poles.extend(getattr(self, f'{stage.lower()}_driver_'
                                               f'compensated_p_{quadrant.lower()}'))

                # Convert lists to numpy arrays
                zeros = np.asarray(zeros)
                poles = np.asarray(poles)

                # Gain is the ratio of poles / zeros
                # NB: if both are empty, then this returns a gain of 1.0
                gain = np.prod(2.0*np.pi*poles) / np.prod(2.0*np.pi*zeros)

                # LTI model
                model = signal.ZerosPolesGain(
                    -2.0*np.pi*zeros,
                    -2.0*np.pi*poles,
                    gain)

                # The frequency response of the LTI model saved to
                # out[stage][quadrant]
                tf = signal.freqresp(model, 2.0*np.pi*frequencies)[1]
                out[stage][quadrant] = tf

        return out

    def uim_dc_gain_Apct(self):
        """This computes the UIM DC gain in units of amps / count

        """

        return 4 * 0.25 * self.dac_gain * self.uim_driver_dc_trans_apv

    def pum_dc_gain_Apct(self):
        """This computes the PUM DC gain in units of amps / count

        """

        return 4 * 0.25 * self.dac_gain * self.pum_driver_dc_trans_apv * \
            self.pum_coil_outf_signflip

    def tst_dc_gain_V2pct(self):
        """This computes the TST DC gain in units of volts**2 / count

        """

        return 4 * 0.25 * self.dac_gain * self.tst_driver_dc_gain_vpv_lv * \
            2 * self.tst_driver_dc_gain_vpv_hv * \
            self.actuation_esd_bias_voltage

    def uim_dc_gain_Npct(self):
        """This computes the UIM DC gain in units of Newtons / count

        """

        return self.uim_dc_gain_Apct() * self.uim_npa

    def pum_dc_gain_Npct(self):
        """
        This computes the PUM DC gain in units of Newtons / count
        """

        return self.pum_dc_gain_Apct() * self.pum_npa

    def tst_dc_gain_Npct(self):
        """This computes the TST DC gain in units of Newtons / count

        """

        return self.tst_dc_gain_V2pct() * self.tst_npv2

    def sus_digital_filters(self, frequencies):
        """
        The transfer function of the SUS digital filters

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        tst_isc_inf_filter_response : `complex128`, array-like
            transfer function response of the ISC_INF digital filters in the
            TST path
        tst_lock_filter_response : `complex128`, array-like
            transfer function response of the LOCK digital filters in the
            TST path
        tst_drive_align_filter_response : `complex128`, array-like
            transfer function response of the DRIVEALIGN digital filters in
            the TST path
        pum_lock_filter_response : `complex128`, array-like
            transfer function response of the LOCK digital filters in the
            PUM path
        pum_drive_align_filter_response : `complex128`, array-like
            transfer function response of the DRIVEALIGN digital filters in the
             PUM path
        uim_lock_filter_response : `complex128`, array-like
            transfer function response of the LOCK digital filters in the UIM
            path
        uim_drive_align_filter_response : `complex128`, array-like
            transfer function response of the DRIVEALIGN digital filters in the
            UIM path

        """

        # Start with TST stage filters
        tst_isc_inf_filter_response, pfilt = \
            compute_digital_filter_response(
                self.dpath(self.sus_filter_file),
                self.tst_isc_inf_bank, self.tst_isc_inf_modules[0],
                self.tst_isc_inf_gain, frequencies)
        tst_lock_filter_response = \
            compute_digital_filter_response(
                self.dpath(self.sus_filter_file),
                self.tst_lock_bank, self.tst_lock_modules[0],
                self.tst_lock_gain, frequencies, pfilt)[0]
        tst_drive_align_filter_response = \
            compute_digital_filter_response(
                self.dpath(self.sus_filter_file),
                self.tst_drive_align_bank, self.tst_drive_align_modules[0],
                self.tst_drive_align_gain, frequencies, pfilt)[0]

        # The PUM stage feedback is to the same SUS as TST stage, so
        # we can reuse the pfilt variable from above
        pum_lock_filter_response = \
            compute_digital_filter_response(
                self.dpath(self.sus_filter_file),
                self.pum_lock_bank, self.pum_lock_modules[0],
                self.pum_lock_gain, frequencies, pfilt)[0]
        pum_drive_align_filter_response = \
            compute_digital_filter_response(
                self.dpath(self.sus_filter_file),
                self.pum_drive_align_bank, self.pum_drive_align_modules[0],
                self.pum_drive_align_gain, frequencies, pfilt)[0]

        # The UIM stage feedback is to the same SUS as PUM stage, so
        # we can reuse the pfilt variable from above
        uim_lock_filter_response = \
            compute_digital_filter_response(
                self.dpath(self.sus_filter_file),
                self.uim_lock_bank, self.uim_lock_modules[0],
                self.uim_lock_gain, frequencies, pfilt)[0]
        uim_drive_align_filter_response = \
            compute_digital_filter_response(
                self.dpath(self.sus_filter_file),
                self.uim_drive_align_bank, self.uim_drive_align_modules[0],
                self.uim_drive_align_gain, frequencies, pfilt)[0]

        return tst_isc_inf_filter_response, tst_lock_filter_response, \
            tst_drive_align_filter_response, pum_lock_filter_response, \
            pum_drive_align_filter_response, uim_lock_filter_response, \
            uim_drive_align_filter_response

    def sus_digital_filters_response(self, frequencies):
        """
        The transfer function of the SUS digital filters

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        uim_digital_filter_response : `complex128`, array-like
            transfer function response of the digital filters in the UIM path
        pum_digital_filter_response : `complex128`, array-like
            transfer function response of the digital filters in the PUM path
        tst_digital_filter_response : `complex128`, array-like
            transfer function response of the digital filters in the TST path

        """

        use_index = False
        idx = 0
        for i, saved_freq_array in enumerate(self.filter_response_freq_list):
            if np.atleast_1d(np.atleast_1d(frequencies) == np.atleast_1d(saved_freq_array)).all():
                idx = i
                use_index = True

        if use_index:
            tst_digital_filter_response = self.tst_digital_filter_response_list[idx]
            pum_digital_filter_response = self.pum_digital_filter_response_list[idx]
            uim_digital_filter_response = self.uim_digital_filter_response_list[idx]

        else:
            tst_isc_inf_filter_response, tst_lock_filter_response, \
                tst_drive_align_filter_response, pum_lock_filter_response, \
                pum_drive_align_filter_response, uim_lock_filter_response, \
                uim_drive_align_filter_response = \
                self.sus_digital_filters(frequencies)

            # Start with TST stage filters
            tst_digital_filter_response = tst_isc_inf_filter_response * \
                tst_lock_filter_response * tst_drive_align_filter_response

            pum_digital_filter_response = tst_isc_inf_filter_response * \
                tst_lock_filter_response * pum_lock_filter_response * \
                pum_drive_align_filter_response

            uim_digital_filter_response = tst_isc_inf_filter_response * \
                tst_lock_filter_response * pum_lock_filter_response * \
                uim_lock_filter_response * uim_drive_align_filter_response

            self.filter_response_freq_list.append(frequencies)
            self.tst_digital_filter_response_list.append(tst_digital_filter_response)
            self.pum_digital_filter_response_list.append(pum_digital_filter_response)
            self.uim_digital_filter_response_list.append(uim_digital_filter_response)

        return uim_digital_filter_response, \
            pum_digital_filter_response, \
            tst_digital_filter_response

    def sus_digital_compensation_response(self, frequencies):
        """
        The transfer function of the SUS compensation filters for driver
        electronics. By default, the output will be a dictionary for UIM,
        PUM, and TST with dictionaries for each containing the quadrants
        UL, LL, UR, LR. Transfer functions are unity with zero phase unless
        the params file/string contains:
        <uim|pum|tst>_compensation_filter_bank_<ul|ll|ur|lr> with a name;
        <uim|pum|tst>_compensation_filter_modules_in_use_<ul|ll|ur|lr> with
        list of modules turned on; and
        <uim|pum|tst>_compensation_filter_gain_<ul|ll|ur|lr>

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        out : dict, `complex128`, array-like
            transfer function response of the digital compensation filters in
            each path and quadrant

        """

        out = {'UIM': {'UL': np.ones(len(frequencies), dtype='complex128'),
                       'LL': np.ones(len(frequencies), dtype='complex128'),
                       'UR': np.ones(len(frequencies), dtype='complex128'),
                       'LR': np.ones(len(frequencies), dtype='complex128')},
               'PUM': {'UL': np.ones(len(frequencies), dtype='complex128'),
                       'LL': np.ones(len(frequencies), dtype='complex128'),
                       'UR': np.ones(len(frequencies), dtype='complex128'),
                       'LR': np.ones(len(frequencies), dtype='complex128')},
               'TST': {'UL': np.ones(len(frequencies), dtype='complex128'),
                       'LL': np.ones(len(frequencies), dtype='complex128'),
                       'UR': np.ones(len(frequencies), dtype='complex128'),
                       'LR': np.ones(len(frequencies), dtype='complex128')}}

        pfilt = None
        for i, stage in enumerate(out.keys()):
            for j, quadrant in enumerate(out[stage].keys()):

                # Emit error if trying to use the old format
                if ((hasattr(self, f'{stage.lower()}_compensation_filter_bank_'
                                   f'{quadrant.lower()}') and
                     not hasattr(self, f'{stage.lower()}_filter_bank_'
                                       f'{quadrant.lower()}')) or
                    (hasattr(self, f'{stage.lower()}_compensation_filter_'
                                   f'modules_in_use_{quadrant.lower()}') and
                     not hasattr(self, f'{stage.lower()}_compensation_filter_'
                                       f'modules_{quadrant.lower()}')) or
                    (hasattr(self, f'{stage.lower()}_compensation_filter_gain_'
                                   f'{quadrant.lower()}') and
                     not hasattr(self, f'{stage.lower()}_filter_gain_{quadrant.lower()}'))):
                    raise KeyError(
                        'Using the old format for driver compensation'
                        ' parameters. Please update your configuration'
                        ' file/string to use the new format.')

                if hasattr(self, f'{stage.lower()}_filter_bank_{quadrant.lower()}'):
                    bank = getattr(self,
                                   f'{stage.lower()}_filter_bank_{quadrant.lower()}')
                if (hasattr(self, f'{stage.lower()}_compensation_filter_'
                                  f'modules_{quadrant.lower()}') and
                        getattr(self, f'{stage.lower()}_front_end_driver_'
                                      f'compensation_{quadrant.lower()}') == 'OFF'):
                    modules = getattr(
                        self,
                        f'{stage.lower()}_compensation_filter_modules_{quadrant.lower()}')
                if hasattr(self, f'{stage.lower()}_filter_gain_{quadrant.lower()}'):
                    gain = getattr(
                        self, f'{stage.lower()}_filter_gain_{quadrant.lower()}')

                if ('bank' in locals() and len(bank) > 0 and
                        'modules' in locals() and 'gain' in locals()):
                    tf, pfilt = compute_digital_filter_response(
                        self.dpath(self.sus_filter_file),
                        bank, modules, gain, frequencies, pfilt)

                    out[stage][quadrant] *= tf
                    del bank, gain, modules

        return out

    def combine_sus_quadrants(self, frequencies):
        """
        Combine the digital and analog paths for the SUS quadrants on a
        per-stage and per-path basis. Meaning
        (0.25 x UIM UL digital x UIM UL analog +
        0.25 x UIM LL digital x UIM LL analog + ...) and so on for each
        stage

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        uim : dict, `complex128`, array-like
            transfer function response of the digital compensation filters in
            each path and quadrant for the UIM
        pum : dict, `complex128`, array-like
            transfer function response of the digital compensation filters in
            each path and quadrant for the PUM
        tst : dict, `complex128`, array-like
            transfer function response of the digital compensation filters in
            each path and quadrant for the TST

        """

        sus_dict_digital = self.sus_digital_compensation_response(frequencies)
        sus_dict_analog = self.analog_driver_response(frequencies)

        uim = 0.25 * (
            sus_dict_digital['UIM']['UL'] * sus_dict_analog['UIM']['UL'] +
            sus_dict_digital['UIM']['LL'] * sus_dict_analog['UIM']['LL'] +
            sus_dict_digital['UIM']['UR'] * sus_dict_analog['UIM']['UR'] +
            sus_dict_digital['UIM']['LR'] * sus_dict_analog['UIM']['LR'])
        pum = 0.25 * (
            sus_dict_digital['PUM']['UL'] * sus_dict_analog['PUM']['UL'] +
            sus_dict_digital['PUM']['LL'] * sus_dict_analog['PUM']['LL'] +
            sus_dict_digital['PUM']['UR'] * sus_dict_analog['PUM']['UR'] +
            sus_dict_digital['PUM']['LR'] * sus_dict_analog['PUM']['LR'])
        tst = 0.25 * (
            sus_dict_digital['TST']['UL'] * sus_dict_analog['TST']['UL'] +
            sus_dict_digital['TST']['LL'] * sus_dict_analog['TST']['LL'] +
            sus_dict_digital['TST']['UR'] * sus_dict_analog['TST']['UR'] +
            sus_dict_digital['TST']['LR'] * sus_dict_analog['TST']['LR'])

        return uim, pum, tst

    def actuation_stage_residual(self, frequencies, stage='TST'):
        """
        This is the actuation residual for a given stage on a given arm (this
        object), meaning it contains everything in the actuation except for
        the output matrix, the digital distribution filters, the actuator
        gain, and suspension dynamics.

        See T1900169

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response
        stage : `str`, optional
            UIM, PUM or TST (default = 'TST')

        Returns
        -------
        A_mn_res : `complex128`, array-like
            transfer function response of the residual actuation function

        """

        # SUS to IOP and IOP to analog digital time delays (see G1601472)
        sus_to_iop_delay_response = signal.dfreqresp(
            digital_delay_filter(1, 16384), 2.0*np.pi*frequencies/16384)[1]
        iop_to_analog_delay_response = signal.dfreqresp(
            digital_delay_filter(4, 65536), 2.0*np.pi*frequencies/65536)[1]

        # digital anti-imaging filter
        digital_ai_filter_response = \
            self.digital_aa_or_ai_filter_response(frequencies)

        # analog anti-imaging filter response
        analog_ai_filter_response = \
            self.analog_aa_or_ai_filter_response(frequencies)

        # Combine analog driver electronics for each quadrant with the
        # digital compensation for each quadrant (if specified), and
        # then take the mean of the four quadrants for each stage. Most
        # often, this only contains the high frequency response that we
        # cannot compensate.
        # Order: [UIM, PUM, TST]
        stages_tfs = self.combine_sus_quadrants(frequencies)
        stage_index = ['uim', 'pum', 'tst'].index(stage.lower())
        stage_tf = stages_tfs[stage_index]

        # Unknown overall time delay
        unknown_actuation_delay = np.exp(-2.0*np.pi*1j *
                                         self.unknown_actuation_delay *
                                         frequencies)

        # Independent time delay
        if (hasattr(self, f'{stage.lower()}_delay') and
                getattr(self, f'{stage.lower()}_delay') != ''):
            delay = getattr(self, f'{stage.lower()}_delay')
            stage_delay = np.exp(-2.0*np.pi*1j*delay*frequencies)
        else:
            stage_delay = 1

        A_mn_res = (sus_to_iop_delay_response *
                    iop_to_analog_delay_response *
                    digital_ai_filter_response *
                    analog_ai_filter_response *
                    stage_tf *
                    unknown_actuation_delay *
                    stage_delay)

        return A_mn_res

    @staticmethod
    def matlab_force2length_response(suspension_file, frequencies):
        """
        Load the ZPK output from the matlab exported values and output a
        frequency response. This function expects variables in a .mat file to
        be UIMz, UIMp, UIMk, PUMz, PUMp, PUMk, TSTz, TSTp, and TSTk

        Parameters
        ----------
        suspension_file: `str`
            path and filename to .mat file that has the ZPK arrays
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        UIM_F2L_freqresp : `complex128`, array-like
            transfer function response of the UIM stage
        PUM_F2L_freqresp : `complex128`, array-like
            transfer function response of the PUM stage
        TST_F2L_freqresp : `complex128`, array-like
            transfer function response of the TST stage

        """

        mat = io.loadmat(suspension_file)
        UIM_F2L = signal.ZerosPolesGain(mat['UIMz'][:, 0], mat['UIMp'][:, 0],
                                        mat['UIMk'][0, 0])
        PUM_F2L = signal.ZerosPolesGain(mat['PUMz'][:, 0], mat['PUMp'][:, 0],
                                        mat['PUMk'][0, 0])
        TST_F2L = signal.ZerosPolesGain(mat['TSTz'][:, 0], mat['TSTp'][:, 0],
                                        mat['TSTk'][0, 0])
        UIM_F2L_freqresp = freqrespZPK(UIM_F2L, 2.0*np.pi*frequencies)
        PUM_F2L_freqresp = freqrespZPK(PUM_F2L, 2.0*np.pi*frequencies)
        TST_F2L_freqresp = freqrespZPK(TST_F2L, 2.0*np.pi*frequencies)

        return UIM_F2L_freqresp, PUM_F2L_freqresp, TST_F2L_freqresp

    def digital_out_to_displacement(self, frequencies):
        """
        This computes the transfer function from the output of the DRIVEALIGN
        filter bank of each stage to displacement of the test mass

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        uim_response : `complex128`, array-like
            transfer function response of the UIM stage
        pum_response : `complex128`, array-like
            transfer function response of the PUM stage
        tst_response : `complex128`, array-like
            transfer function response of the TST stage

        """

        # Analog force to length response for each stage from Matlab
        sus_file = self.dpath(self.suspension_file)
        [uim_f2l_response,
         pum_f2l_response,
         tst_f2l_response] = self.matlab_force2length_response(sus_file,
                                                               frequencies)

        # Actuation stage residuals
        A_u_res = self.actuation_stage_residual(frequencies, stage='UIM')
        A_p_res = self.actuation_stage_residual(frequencies, stage='PUM')
        A_t_res = self.actuation_stage_residual(frequencies, stage='TST')

        # Everything together now!
        tst_response = (A_t_res *
                        self.tst_dc_gain_Npct() *
                        tst_f2l_response)
        pum_response = (A_u_res *
                        self.pum_dc_gain_Npct() *
                        pum_f2l_response)
        uim_response = (A_p_res *
                        self.uim_dc_gain_Npct() *
                        uim_f2l_response)

        return uim_response, pum_response, tst_response

    def drivealign_to_longitudinal_displacement(self, frequencies):
        """
        This computes the transfer function from the input of the DRIVEALIGN
        filter bank of each stage to displacement of the test mass
        where "longitudinal" is defined as in the traditional SUS
        Euler Basis: positive longitudinal displacement is along a
        vector perpendicular to the plane of the HR surface
        (positive ETMX displacement is in the -X global IFO direction,
        positive ITMY displacement is in the +Y global IFO direction).

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        uim_response : `complex128`, array-like
            transfer function response of the UIM stage
        pum_response : `complex128`, array-like
            transfer function response of the PUM stage
        tst_response : `complex128`, array-like
            transfer function response of the TST stage

        """

        # Digital out to displacement of the TST stage by each of the
        # SUS stages
        [uim_response,
         pum_response,
         tst_response] = self.digital_out_to_displacement(frequencies)

        # Need to add in the DRIVEALIGN bank for each stage
        # Like before, we can reuse the pfilt variable if the preceeding
        # stage feedback goes to the same SUS
        tst_drive_align_filter_response, pfilt = \
            compute_digital_filter_response(
                self.dpath(self.sus_filter_file),
                self.tst_drive_align_bank, self.tst_drive_align_modules[0],
                self.tst_drive_align_gain, frequencies)
        pum_drive_align_filter_response = \
            compute_digital_filter_response(
                self.dpath(self.sus_filter_file),
                self.pum_drive_align_bank, self.pum_drive_align_modules[0],
                self.pum_drive_align_gain, frequencies, pfilt)[0]
        uim_drive_align_filter_response = \
            compute_digital_filter_response(
                self.dpath(self.sus_filter_file),
                self.uim_drive_align_bank, self.uim_drive_align_modules[0],
                self.uim_drive_align_gain, frequencies)[0]

        uim_response *= uim_drive_align_filter_response
        pum_response *= pum_drive_align_filter_response
        tst_response *= tst_drive_align_filter_response

        return uim_response, pum_response, tst_response

    def drivealign_to_darm_displacement(self, frequencies):
        """
        This computes the transfer function from the input of the DRIVEALIGN
        filter bank of each action stage to DARM by accounting for how
        positive longitudinal drive for each actuation stage pushes the given
        test mass to creates DARM.

        DARM displacement is defined by

        .. math ::

            + DARM = + Delta L \\
            = (Delta L_X - Delta L_Y) \\
            = - (Delta L_EX + Delta L_IX) + (Delta L_EY + Delta L_IY)

        This is particularly useful for computing EPICS records

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        uim_response : `complex128`, array-like
            transfer function response of the UIM stage
        pum_response : `complex128`, array-like
            transfer function response of the PUM stage
        tst_response : `complex128`, array-like
            transfer function response of the TST stage

        """

        longitudinal_displacement = (
            self.drivealign_to_longitudinal_displacement(frequencies))

        # Multiply each stage by the DARM feedback sign
        for idx, tf in enumerate(longitudinal_displacement):
            tf *= self.darm_feedback_sign
        darm_displacement = longitudinal_displacement  # rename after DARM sign

        # Return the DARM displacement tuple
        return darm_displacement

    def known_actuation_terms_for_measurement(self, frequencies):
        """
        This method computes all known terms which are divided out of the
        DARM_ERR / SUS_EXC full IFO actuation measurement for each stage
        such that the remaining dimensional actuation strength, H_Ai, can
        be fit by MCMC.

        This computes the transfer function from the input of the DRIVEALIGN
        filter bank (same as the SUS_EXC point) of each stage to DARM
        displacement from any test mass, including everything except the
        N/(drive unit); either amps or volts**2, for which the MCMC will fit.
        The DARM feedback sign of the arm the measurement is taken on since
        we read out the measurement with DARM_ERR.

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response

        Returns
        -------
        uim_response : `complex128`, array-like
            transfer function response of the UIM stage, everything but N/ct
        pum_response : `complex128`, array-like
            transfer function response of the PUM stage, everything but N/ct
        tst_response : `complex128`, array-like
            transfer function response of the TST stage, everything but N/ct

        """

        [uim_response,
         pum_response,
         tst_response] = self.drivealign_to_darm_displacement(frequencies)

        uim_response *= (self.uim_dc_gain_Apct() / self.uim_dc_gain_Npct())
        pum_response *= (self.pum_dc_gain_Apct() / self.pum_dc_gain_Npct())
        tst_response *= (self.tst_dc_gain_V2pct() / self.tst_dc_gain_Npct())

        return uim_response, pum_response, tst_response

    def compute_actuation_single_stage(self, frequencies, stage='TST'):
        """
        Compute the actuation function transfer function for a single stage
        (see G1501518). This transfer function is from the DARM control
        input at the SUS user model, or the input to the ISCINF bank, to
        DARM displacement.

        This does not include the OMC to SUS model jump delay

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response
        stage : `str`, optional
            UIM, PUM or TST (default = 'TST')

        Returns
        -------
        tf : `complex128`, array-like
            transfer function response of the actuation function

        """

        # Digital output to TST displacement
        # Note that no feedback or output matrix values are applied here
        # Note also that digital_out_to_displacement does not have the
        # OMC to SUS delay
        [uim_response,
         pum_response,
         tst_response] = self.digital_out_to_displacement(frequencies)

        # All SUS digital filters
        [uim_digital_filter_response,
         pum_digital_filter_response,
         tst_digital_filter_response] = self.sus_digital_filters_response(
            frequencies)

        # Choose the right stage
        if stage in ['UIM', 'L1']:
            sus_response = uim_response * uim_digital_filter_response
        elif stage in ['PUM', 'L2']:
            sus_response = pum_response * pum_digital_filter_response
        elif stage in ['TST', 'L3']:
            sus_response = tst_response * tst_digital_filter_response
        else:
            raise ValueError('stage must be UIM, PUM, TST, L1, L2, or L3')

        sus_response *= self.darm_feedback_sign

        # Everything *above* here is modelling the physical actuator that we
        # have. However, the calibration group's convention is that A in their
        # loop drawings is minus the physical actuator (which allows the
        # explicit writing of the minus sign in the loop drawing to indicate
        # negative feedback). Ideally we'd just put
        # A = -1 * (TOP + UIM + PUM + TST),
        # but the actuation products below (which are used for other things
        # downstream) are expecting to receive minus of each stage of the
        # physical actuator. So, we multiply every stage by -1, and then sum
        # them up, and the result will be that A is negative of the physical
        # actuator. JCD_26Mar2019
        # See more details in T1800456.
        sus_response *= -1

        return sus_response


class DARMActuationModel(Model):
    """
    An DARM actuation model object

    This is a class to set up the model for the DARM loop from a
    configuration file with all the information about where the data is stored

    """

    def __init__(self, config):
        super().__init__(config, measurement='actuation')
        self.xarm = self.yarm = None
        if 'actuation_x_arm' in self._config:
            self.xarm = ActuationModel(config, measurement='actuation_x_arm')
        if 'actuation_y_arm' in self._config:
            self.yarm = ActuationModel(config, measurement='actuation_y_arm')

    def darm_output_matrix_values(self):
        """Turns the output matrix values into an array for the QUAD SUS

        """

        output_matrix = np.zeros((2, 4))
        for m in range(len(output_matrix[:, 0])):
            for n in range(len(output_matrix[0, :])):
                if m == 0 and self.darm_feedback_x[n] == 'ON':
                    output_matrix[m, n] = self.darm_output_matrix[0]
                elif m == 1 and self.darm_feedback_y[n] == 'ON':
                    output_matrix[m, n] = self.darm_output_matrix[1]

        return output_matrix

    def stage_super_actuator(self, frequencies, stage='TST', syserr_dict=None):
        """
        Compute the super actuator transfer function for a specific stage.
        In this case, the "stage super actuator" is created by choosing a
        specific stage and then for each QUAD, the stage transfer function to
        DARM is summed together.

        This transfer function is from DARM_CTRL to meters sensed by the IFO.
        Note that the sign of the DARM_ERR signal is dependent upon which arm
        is under control.

        The OMC to SUS delay is included in this transfer function.

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response
        stage : `str`, optional
            SUS stage for the calculation, 'UIM', 'PUM', or 'TST'
        syserr_dict : `dict`, optional
            dict of multiplicative values, ex.:
            {'xarm': {'UIM': `complex`, array-like}}

        Returns
        -------
        super_actuation : `complex128`, array-like
            transfer function response of the actuation function

        """

        output_matrix = self.darm_output_matrix_values()

        super_actuation = np.zeros(len(frequencies), dtype='complex128')

        stage_idx = ['top', 'uim', 'pum', 'tst'].index(stage.lower())
        for arm_idx, arm in enumerate(['x', 'y']):
            if output_matrix[arm_idx, stage_idx] != 0.0:
                assert stage_idx > 0, 'top has not been implemented'

                if hasattr(self, f'{arm}arm') is False:
                    raise ValueError(
                        'Must provide an ActuationModel object')
                else:
                    this_arm = getattr(self, f'{arm}arm')

                single_stage = this_arm.compute_actuation_single_stage(
                        frequencies, stage=stage.upper())

                if (syserr_dict is not None and
                        ['xarm', 'yarm'][arm_idx] in syserr_dict and
                        stage.upper() in syserr_dict[['xarm', 'yarm'][arm_idx]]):
                    single_stage *= (
                        syserr_dict[['xarm', 'yarm'][arm_idx]][stage.upper()])

                super_actuation += (
                    output_matrix[arm_idx, stage_idx] * single_stage)

        # OMC to SUS model jump delay, sending DARM_CTRL from the omc
        # front-end model via IPC to any QUAD SUS model, identical for any
        # QUAD, either at the corner or end.
        # Note that SUS to IOP, IOP to analog, and unknown delays have already
        # been applied in compute_actuation_single_stage.
        # For measurement confirmation of computational time delays see
        # G1601472.
        omc_to_sus_delay_response = signal.dfreqresp(
            digital_delay_filter(1, 16384), 2.0*np.pi*frequencies/16384)[1]
        super_actuation *= omc_to_sus_delay_response

        return super_actuation

    def stage_super_actuator_drivealign(self, frequencies, stage='TST',
                                        syserr_dict=None):
        """
        Compute the super actuator transfer function for a specific stage.
        In this case, the "stage super actuator" is created by choosing a
        specific stage and then for each QUAD, the stage transfer function to
        DARM is summed together.

        This transfer function is from the input of the DRIVEALIGN filter bank
        to meters sensed by the IFO. This method assumes that the signal
        feedback to a single stage would be positive, but if the signal is sent
        to both arms (like a DARM signal) then X and Y are 180 degrees out of
        phase.

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response
        stage : `str`
            SUS stage for the calculation, 'UIM', 'PUM', or 'TST'
        syserr_dict : `dict`, optional
            dict of multiplicative values, ex.:
            {'xarm': {'UIM': `complex`, array-like}}

        Returns
        -------
        super_actuation : `complex128`, array-like
            transfer function response of the actuation function
        """

        assert stage.lower() != 'top', 'top has not been implemented'

        # Get the output matrix values
        output_matrix = self.darm_output_matrix_values()

        # Get the stage index
        stage_idx = ['top', 'uim', 'pum', 'tst'].index(stage.lower())

        # We have two scenarios:
        # 1) Both arms are used for servo control. X and Y arms receive the
        # same signal but 180 degrees out of phase, so here we end up
        # SUBTRACTING the Y arm from X arm due to this 180 degrees difference
        # 2) Only one arm uses feedback control. We want to maintain the
        # historical infrastructure for single-arm feedback, so we don't
        # need to account for any signs when combining arm contributions.
        # There may be a cleaner way to code this...
        if output_matrix[0, stage_idx] != 0 and output_matrix[1, stage_idx] != 0:
            # Check that we have attributes for both arms
            if hasattr(self, 'xarm') is False or hasattr(self, 'yarm') is False:
                raise ValueError('Must provide an ActuationModel object for both arms')

            xarm_contrib = (
                self.xarm.drivealign_to_darm_displacement(frequencies)[stage_idx-1])
            yarm_contrib = (
                self.yarm.drivealign_to_darm_displacement(frequencies)[stage_idx-1])

            if syserr_dict is not None:
                if 'xarm' in syserr_dict and stage.upper() in syserr_dict['xarm']:
                    xarm_contrib *= syserr_dict['xarm'][stage.upper()]
                if 'yarm' in syserr_dict and stage.upper() in syserr_dict['yarm']:
                    yarm_contrib *= syserr_dict['yarm'][stage.upper()]

            super_actuation = (xarm_contrib - yarm_contrib)
        else:
            for arm_idx, arm in enumerate(['x', 'y']):
                if output_matrix[arm_idx, stage_idx] != 0.0:
                    if hasattr(self, f'{arm}arm') is False:
                        raise ValueError('Must provide an ActuationModel object')
                    else:
                        this_arm = getattr(self, f'{arm}arm')

                    super_actuation = this_arm.drivealign_to_darm_displacement(
                        frequencies)[stage_idx-1]

                    if (syserr_dict is not None and
                            f'{arm}arm' in syserr_dict and
                            stage.upper() in syserr_dict[f'{arm}arm']):
                        super_actuation *= syserr_dict[f'{arm}arm'][stage.upper()]

        return super_actuation

    def arm_super_actuator(self, frequencies, arm='x', syserr_dict=None):
        """
        Compute the super actuator transfer function for a specific arm.
        In this case, the "arm super actuator" is created by choosing a
        specific arm and then for each QUAD, the arm transfer function to
        DARM is summed together.

        This transfer function is from DARM_CTRL to meters sensed by the IFO.
        Note that the sign of the DARM_ERR signal is dependent upon which arm
        is under control.

        The OMC to SUS delay is included in this transfer function.

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response
        arm : `str`
            arm for the calculation, 'x' or 'y'
        syserr_dict : `dict`, optional
            dict of multiplicative values, ex.:
            {'xarm': {'UIM': `complex`, array-like}}

        Returns
        -------
        super_actuation : `complex128`, array-like
            transfer function response of the actuation function

        """

        output_matrix = self.darm_output_matrix_values()

        super_actuation = np.zeros(len(frequencies), dtype='complex128')

        arm_idx = ['x', 'y'].index(arm.lower())
        for stage_idx, stage in enumerate(['top', 'uim', 'pum', 'tst']):
            if output_matrix[arm_idx, stage_idx] != 0.0:
                assert stage_idx > 0, 'top has not been implemented'

                if hasattr(self, f'{arm}arm') is False:
                    raise ValueError(
                        'Must provide an ActuationModel object')
                else:
                    this_arm = getattr(self, f'{arm}arm')

                single_stage = this_arm.compute_actuation_single_stage(
                        frequencies, stage=stage.upper())

                if (syserr_dict is not None and
                        ['xarm', 'yarm'][arm_idx] in syserr_dict and
                        stage.label() in syserr_dict[['xarm', 'yarm'][arm_idx]]):
                    single_stage *= (
                        syserr_dict[['xarm', 'yarm'][arm_idx]][stage.upper()])

                super_actuation += (
                    output_matrix[arm_idx, stage_idx] * single_stage)

        # OMC to SUS model jump delay, sending DARM_CTRL from the omc
        # front-end model via IPC to any QUAD SUS model, identical for any
        # QUAD, either at the corner or end.
        # Note that SUS to IOP, IOP to analog, and unknown delays have already
        # been applied in compute_actuation_single_stage.
        # For measurement confirmation of computational time delays see
        # G1601472.
        omc_to_sus_delay_response = signal.dfreqresp(
            digital_delay_filter(1, 16384), 2.0*np.pi*frequencies/16384)[1]
        super_actuation *= omc_to_sus_delay_response

        return super_actuation

    def compute_actuation(self, frequencies, syserr_dict=None):
        """
        Compute the entire actuation function transfer function (see G1501518).
        This transfer function is from DARM_CTRL to meters sensed by the IFO.
        Note that the sign of the DARM_ERR signal is dependent upon which arm
        is under control.

        Parameters
        ----------
        frequencies : `float`, array-like
            array of frequencies to compute the response
        syserr_dict : `dict`, optional
            dict of multiplicative values, ex.:
            {'xarm': {'UIM': `complex`, array-like}}

        Returns
        -------
        tf : `complex128`, array-like
            transfer function response of the actuation function

        """

        darm_actuation = np.zeros(len(frequencies), dtype='complex128')

        # This has the OMC to SUS time delay
        for stage_idx, stage in enumerate(['top', 'uim', 'pum', 'tst']):
            darm_actuation += self.stage_super_actuator(
                frequencies, stage=stage.upper(), syserr_dict=syserr_dict)

        return darm_actuation



import unittest
import pydarm
import numpy as np


class TestDigitalAAFilterResponse(unittest.TestCase):

    def setUp(self):
        # Pre-computed values
        # anti_aliasing_rate_string = 16k
        # anti_aliasing_method      = biquad
        # frequencies = np.logspace(0, np.log10(5000.), 10)
        self.frequencies = np.logspace(0, np.log10(5000.), 10)
        self.known_digital_aa_filter_response = np.array(
            [1.0000006743118521-0.00026859554703262606j,
             1.0000005577918698-0.0006919831728579482j,
             0.9999997844069123-0.0017827587938075234j,
             0.9999946510415249-0.004592951080196375j,
             0.9999605723740017-0.011833290098468159j,
             0.9997340799606691-0.03049404992719877j,
             0.9982174966256611-0.07869700811767119j,
             0.9875624565339749-0.20500805282580065j,
             0.8902071582110597-0.5598509390584541j,
             -0.3367135303069774-1.013081908852281j])

    def tearDown(self):
        del self.frequencies
        del self.known_digital_aa_filter_response

    def test_digital_aa_filter_response(self):
        """ Test the digital AA filter response """
        C = pydarm.sensing.SensingModel('''
[metadata]
[interferometer]
[sensing]
anti_aliasing_rate_string = 16k
anti_aliasing_method = biquad
''')
        digital_aa_filter_response = C.digital_aa_or_ai_filter_response(
            self.frequencies)
        for n in range(len(self.frequencies)):
            self.assertAlmostEqual(
                np.abs(digital_aa_filter_response[n]),
                np.abs(self.known_digital_aa_filter_response[n]))
            self.assertAlmostEqual(
                np.angle(digital_aa_filter_response[n], deg=True),
                np.angle(self.known_digital_aa_filter_response[n], deg=True))


class TestDigitalAIFilterResponse(unittest.TestCase):

    def setUp(self):
        # Pre-computed values
        # anti_imaging_rate_string = 16k
        # anti_imaging_method      = biquad
        # frequencies = np.logspace(0, np.log10(5000.), 10)
        self.frequencies = np.logspace(0, np.log10(5000.), 10)
        self.known_digital_ai_filter_response = np.array(
            [1.0000006743118521-0.00026859554703262606j,
             1.0000005577918698-0.0006919831728579482j,
             0.9999997844069123-0.0017827587938075234j,
             0.9999946510415249-0.004592951080196375j,
             0.9999605723740017-0.011833290098468159j,
             0.9997340799606691-0.03049404992719877j,
             0.9982174966256611-0.07869700811767119j,
             0.9875624565339749-0.20500805282580065j,
             0.8902071582110597-0.5598509390584541j,
             -0.3367135303069774-1.013081908852281j])

    def tearDown(self):
        del self.frequencies
        del self.known_digital_ai_filter_response

    def test_digital_ai_filter_response(self):
        """ Test the digital AI filter response """
        A = pydarm.actuation.ActuationModel('''
[metadata]
[interferometer]
[actuation_x_arm]
anti_imaging_rate_string = 16k
anti_imaging_method = biquad
''',
                                            measurement='actuation_x_arm')
        digital_ai_filter_response = A.digital_aa_or_ai_filter_response(
            self.frequencies)
        for n in range(len(self.frequencies)):
            self.assertAlmostEqual(
                np.abs(digital_ai_filter_response[n]),
                np.abs(self.known_digital_ai_filter_response[n]))
            self.assertAlmostEqual(
                np.angle(digital_ai_filter_response[n], deg=True),
                np.angle(self.known_digital_ai_filter_response[n], deg=True))


if __name__ == '__main__':
    unittest.main()
